import subprocess
from tempfile import mkdtemp
from jlp import logger
from textwrap import dedent
import re


class Bazaar(object):
    def _call(self, args):
        try:
            subprocess.check_call(['bzr', ] + args)
        except subprocess.CalledProcessError:
            logger.error('Call for bzr %r failed', args)
            exit(1)

    def _call_output(self, args):
        try:
            output = subprocess.check_output(['bzr', ] + args,
                                             stderr=open('/dev/null', 'w'))
            return output
        except subprocess.CalledProcessError:
            logger.error('Call for bzr %r failed', args)
            exit(1)

    def whoami(self, who):
        self._call(['whoami', who, ])

    def branch(self, branch, target_dir):
        self._call(['branch', branch, target_dir])

    def revert(self, target_dir):
        self._call(['revert', target_dir])

    def merge(self, branch):
        self._call(['merge', branch])

    def build(self, recipe, target):
        self._call(['build', recipe, target])

    def revno(self, target):
        return int(self._call_output(['revno', target]).strip())


class BzrRecipe:
    branch = None
    packagingBranch = None
    buildDir = None
    packageVersion = None
    upstreamVersion = None
    sourceName = None
    revision = None
    packagingRevision = None
    bzr = None
    recipe = None
    recipe_template = dedent('''\
                             # bzr-builder format 0.4
                             {main_branch} {revision}
                             {nest_part_cmd}
                             ''')

    def __init__(self, branch, packagingBranch=None, revision=None):
        if branch == packagingBranch:
            packagingBranch = None

        self.branch = branch
        self.packagingBranch = packagingBranch
        self.revision = revision

    def generateRecipe(self):
        self.bzr = Bazaar()
        if self.revision:
            revision = 'revno:' + str(self.revision)
        else:
            self.revision = self.bzr.revno(self.branch)
            revision = 'revno:' + str(self.revision)

        if self.packagingBranch:
            nest_part_cmd = ('nest-part ubuntu ' +
                             self.packagingBranch +
                             ' debian')
            self.packagingRevision = self.bzr.revno(self.packagingBranch)
        else:
            nest_part_cmd = '# No packaging branch'
            self.packagingRevision = 0

        self.recipe = self.recipe_template.format(main_branch=self.branch,
                                                  nest_part_cmd=nest_part_cmd,
                                                  revision=revision)

    def getTrunkDir(self):
        if self.buildDir:
            return self.buildDir + '/trunk'
        return None

    def getBuildDir(self):
        return self.buildDir

    def buildTree(self):
        if self.buildDir:
            return self.buildDir
        self.generateRecipe()
        directory = mkdtemp()
        self.buildDir = directory
        logger.debug('building bazaar recipe in: ' + directory)
        recipe_path = directory + '/trunk.recipe'
        with open(recipe_path, 'w') as recipe_file:
            recipe_file.write(self.recipe)
        self.bzr.build(recipe_path, self.getTrunkDir())

    def createTarball(self, tarballName):
        self.buildTree()
        tarball = self.buildDir + '/' + tarballName
        logger.debug('going to create: ' + tarball)
        return subprocess.call(["tar", "-czf", tarball, '-C', self.buildDir,
                                'trunk'])

    def getPackagingBranchRevno(self):
        self.buildTree()
        return self.packagingRevision

    def getTrunkRevno(self):
        self.buildTree()
        return self.revision

    def _parseVersion(self, dpkgOutput):
        m = re.search('^Version:\s*(.*)$', dpkgOutput,
                      flags=re.MULTILINE)
        if not m:
            return None
        self.upstreamVersion = m.group(1)
        self.packageVersion = m.group(1)
        m = re.match('(.*)-[0-9]ubuntu.*', self.packageVersion)
        if m:
            self.upstreamVersion = m.group(1)
        logger.debug('UpstreamVersion parsed: ' + self.upstreamVersion)
        logger.debug('PackageVersion parsed: ' + self.packageVersion)
        return self.packageVersion

    def _parseSourceName(self, dpkgOutput):
        m = re.search('^Source:\s*(.*)$', dpkgOutput,
                      flags=re.MULTILINE)
        if not m:
            return None
        self.sourceName = m.group(1)
        logger.debug('SourceName parsed: ' + self.sourceName)
        return self.sourceName

    def _parseChangelog(self):
        if self.sourceName and self.packageVersion:
            return True
        parsedChangelog = subprocess.check_output('dpkg-parsechangelog',
                                                  cwd=self.getTrunkDir())
        if not self._parseVersion(parsedChangelog):
            logger.debug('PackageVersion not foud')
            return False
        if not self._parseSourceName(parsedChangelog):
            logger.debug('SourceName not found')
            return False
        return True

    def getPackageVersion(self):
        self.buildTree()
        self._parseChangelog()
        return self.packageVersion

    def getUpstreamVersion(self):
        self.buildTree()
        self._parseChangelog()
        return self.upstreamVersion

    def getSourceName(self):
        self.buildTree()
        self._parseChangelog()
        return self.sourceName

    def revert(self):
        if self.bzr:
            self.bzr.revert(self.getTrunkDir())

    def cleanup(self):
        if self.buildDir:
            subprocess.call(['rm', '-rf', self.getBuildDir()])
        self.buildDir = None
        self.packageVersion = None
        self.upstreamVersion = None
        self.sourceName = None
        self.packagingRevision = None
        self.bzr = None
        self.recipe = None
