from argparse import ArgumentParser
from jlp import (SocketLock, get_launchpad, jenkinsutils, logger,
                 get_config_option)
from jlp.launchpadutils import get_merge_proposals
import os


LAUNCHPADLIB_DIR_PREFIX = os.path.expanduser('~/.launchpadlib')


def trigger_ci_build(branch, job, jenkins_url, launchpadlib_dir, repo_type):
    """Trigger a CI build."""

    launchpad = get_launchpad(launchpadlib_dir)
    mps = get_merge_proposals(launchpad, branch, ['Needs review'], repo_type)
    jenkinsutils.trigger_ci_build(launchpad, mps, job, jenkins_url)


def trigger_al_build(branch, job, jenkins_url, fasttrack_merge,
                     launchpadlib_dir, repo_type):
    """Trigger an autolanding build."""

    launchpad = get_launchpad(launchpadlib_dir)
    mps = get_merge_proposals(launchpad, branch, ['Approved'], repo_type)
    jenkinsutils.trigger_al_build(launchpad, mps, job, jenkins_url,
                                  fasttrack_merge)


def trigger_jenkins(branch,
                    autoland=False,
                    trigger_ci=False,
                    job=None,
                    jenkins_url=None,
                    fasttrack_merge=False,
                    lock_name='',
                    repo_type='auto'):
    """Trigger a Jenkins build."""

    if jenkins_url is None:
        jenkins_url = get_config_option('jenkins_url')

    launchpadlib_dir = '{}-{}'.format(LAUNCHPADLIB_DIR_PREFIX, lock_name)
    logger.debug('Using "{}" as launchpadlib cache directory.'.format(
        launchpadlib_dir))

    if trigger_ci:
        trigger_ci_build(branch, job, jenkins_url, launchpadlib_dir, repo_type)
    elif autoland:
        trigger_al_build(branch, job, jenkins_url, fasttrack_merge,
                         launchpadlib_dir, repo_type)
    return 1


def launchpadTrigger():

    parser = ArgumentParser(
        description="Trigger a Jenkins job resulting from a merge proposal.")
    parser.add_argument('-b', '--branch', required=True,
                        help="the proposed Launchpad branch to merge")
    parser.add_argument('-t', '--trigger-ci',
                        action='store_true',
                        help="trigger a 'CI' build")
    parser.add_argument('-l', '--autoland',
                        action='store_true',
                        help="trigger an autolanding build")
    parser.add_argument('-j', '--job', required=True,
                        help="name of the Jenkins job to trigger")
    parser.add_argument('-e', '--jenkins-url',
                        help="URL of the Jenkins job to trigger")
    parser.add_argument('-f', '--fasttrack-merge',
                        default=False,
                        action='store_true',
                        help='''skip the build process for simple branch
                        changes (i.e. changelog only updates)''')
    parser.add_argument('-L', '--lock-name',
                        default='launchpadTrigger',
                        help='''Name of the lock and launchpadlib cache
                        to be used''')
    parser.add_argument("-r", "--repo_type", default="auto",
                        help="Repository type [bazaar, git, or auto]")

    args = vars(parser.parse_args())

    if not args['trigger_ci'] and not args['autoland']:
        parser.error(
            'You need to specify one of --trigger-ci (-t) or --autoland (-l)')

    # this is here because launchpadlib is not thread safe (yet)
    # (see also lp:459418 and lp:1025153) and also to limit the
    # amount of concurrent connections to launchpad
    trigger_lock_name = '{prefix}-{instance_name}'.format(
        prefix=get_config_option('lock_name'),
        instance_name=args['lock_name'])
    logger.debug('Going to acquire launchpad lock "{}"'
                 ' for launchpadTrigger'.format(trigger_lock_name))
    with SocketLock(trigger_lock_name):
        logger.debug('Lock acquired for launchpadTrigger')
        return trigger_jenkins(**args)
