import re
import os
from xml.dom.minidom import parseString
import jenkins
import urlparse
import urllib2
import launchpadutils
from textwrap import dedent
from . import get_json_jenkins
from lazr.restfulclient.errors import Unauthorized
from jlp import logger, get_config_option


def normalize_url(url):
    return urlparse.urljoin(url,
                            urlparse.urlparse(url).path.replace('//', '/'))


def _actions_has_param(actions, param, value):
    """ Internal method to see if actions json object (as returned by jenkins)
    includes a given parameter (and value).

    It returns True or False based on whether a parameter/value pair is
    included in the given jenkins job actions ("actions" is where jenkins
    stores the parameters)

    :param actions: json result (list of dictionaries) as returned by jenkins
    :param param: name of the jenkins job parameter to check
                  (e.g. 'merge_proposal')
    :param value: value that we want to check

    Returns True/False.
    """
    for action in actions:
        if 'parameters' not in action:
            continue
        for parameter in action['parameters']:
            if parameter['name'] == param and 'value' in parameter and \
                    parameter['value'] == value:
                return True
    return False


def _actions_has_all_params(actions, job_params):
    """ Internal method to see if actions json object (as returned by jenkins)
    includes all given job_params. See also _actions_has_param.

    The main difference between this and _actions_has_param is that this
    method checks for all params in a given dictionary (job_params).

    :param actions: json result (list of dictionaries) as returned by jenkins
    :param job_params: dictionary of parameters to check in the form of:
                       {'parameter': 'value', 'parameter2': 'value2'}

    Returns True/False.
    """
    all_params = True
    for param in job_params:
        all_params = all_params and \
            _actions_has_param(actions, param, job_params[param])
    return all_params


def _deduplicate_builds(seq):
    """ Internal method to remove duplicate builds from a build list"""
    seen = set()
    return [x for x in seq if str(x) not in seen and not seen.add(str(x))]


def get_running_builds(job, job_params={}):
    """ For a given jenkins job return a list of currently active builds
        :param job: name of the jenkins job
        :param job_params: dictionary of job parameters

        Returns an empty list if there is no active build with given
        parameters. If job_params is empty (default) it is assumed no check
        for job_params is required.
    """

    json_jenkins = get_json_jenkins()
    jenkins_url = get_config_option('jenkins_url')
    request = \
        '{}/job/{}/api/json'.format(jenkins_url, job) +\
        '?depth=1&tree=builds[number,building,actions[parameters[name,value]]]'
    data = json_jenkins.get_json_data(request, append_api=False)
    if 'builds' not in data:
        return []
    result = []

    for build in data['builds']:
        if build['building'] and \
                _actions_has_all_params(build['actions'], job_params):
            result.append(build['number'])
    return result


def _get_build_actions(job, build):
    """Internal method. For a given job and build return the "actions" object
    as returned by jenkins. This object holds the list of parameters of a
    given build.

    :param: job: jenkins job name
    :param: build: jenkins build number
    """

    json_jenkins = get_json_jenkins()
    jenkins_url = get_config_option('jenkins_url')

    job = json_jenkins.get_json_data('{}/job/{}/{}'.format(
        jenkins_url, job, build))
    if 'actions' not in job:
        return []
    return job['actions']


def _is_item_queued_by_upstream(item, upstream_job, upstream_params):
    """Internal method to check if a given queued item
    (as returned by jenkins; see the "$jenkins_url/queue/json/api" API) was
    queued by an upstream job with a given upstream_job parameters.

    The idea here is to check e.g. if generic_land (item) was queued by
    unity-autolanding (upstream_job) with {'merge_proposal': 'http://'}
    (upstream_params). If there is such a generic-land queued then this method
    returns True otherwise it returns False

    :param item: "item" as returned in the json jenkins reply (see the queue
                 API)
    :param: upstream_job: name of the upstream_job
    :param upstream_params: dictionary of the upstream job params we are
           interested in
    """
    upstream_project = None
    upstream_build = None

    if 'actions' not in item:
        return False

    for action in item['actions']:
        upstream_project = None
        if 'causes' not in action:
            continue
        for cause in action['causes']:
            if isinstance(cause, dict) and 'upstreamProject' in cause:
                upstream_project = cause['upstreamProject']
                # if the upstreamProject differes we can return False right
                # away
                if upstream_project != upstream_job:
                    return False

                upstream_build = cause['upstreamBuild']
                if upstream_params:
                    upstream_actions = _get_build_actions(
                        upstream_project, upstream_build)
                    if _actions_has_all_params(upstream_actions,
                                               upstream_params):
                        return True
                else:
                    return True
    return False


def is_job_queued(job, queue=None, job_params={}, upstream_job=None,
                  upstream_params={}):
    """ Return if a job is queued in a given jenkins queue.
        If upstream_job is supplied the job in queue must have been triggered
        by upstream_job.
        If upstream_params are given then the triggering upstream_job must
        include the given upstream_params (dictionary of the form of
        {'param1': 'value1'}).

        :param job: jenkins job name
        :param queue: jenkins queue (as returned by $jenkinsurl/queue/api/json)
        :param upstream_job: name of the upstream job that was supposed to
               queue job
        :param upstream_params: parameters to check in the upstream_job
    """
    if queue is None:
        json_jenkins = get_json_jenkins()
        jenkins_url = get_config_option('jenkins_url')
        queue = json_jenkins.get_json_data('{}/queue'.format(jenkins_url))

    for item in queue['items']:
        if 'task' in item and 'name' in item['task']:
            # we can ignore jobs that are different from the job we are
            # looking for
            if item['task']['name'] != job:
                continue

            # check if the required parameters are present
            if job_params and 'actions' in item and \
                    not _actions_has_all_params(item['actions'], job_params):
                continue

            # if we have a match and we don't need to check
            # for upstream_job then we are done
            if upstream_job is None:
                return True

            if _is_item_queued_by_upstream(item,
                                           upstream_job,
                                           upstream_params):
                return True

    return False


def is_downstream_job_running(downstream_project,
                              upstream_job, upstream_params):
    """Check if downstream job is running.
    It returns True only if the downstream_project was triggered by
    upstream_job and upstream_job has upstream_params (e.g. a given merge
    proposal).

    This is unfortunately not 100% reliable. For some reasons the json request
    in (&tree=builds[number,url,...]) sometimes doesn't return what it should
    (maybe the jenkins needs to update some internal state). Once the
    relationship between jobs is somehow recorded the subsequent calls always
    return them correctly.

    :param downstream_project: name of the jenkins downstream_project (e.g.
           generic-land)
    :param upstream_job: name of the upstream_job (e.g. unity-autolanding)
    :param upstream_params: dictionary of the parameters (e.g.
           {'merge_proposal': 'http://...'})
    """
    json_jenkins = get_json_jenkins()
    jenkins_url = get_config_option('jenkins_url')
    json_request = \
        '/api/json?depth=2&tree=builds[number,url,building,' + \
        'actions[causes[upstreamBuild,upstreamProject]]],' + \
        'upstreamProjects[name,builds[number,actions[parameters[name,value]]]]'

    request = '{}job/{}{}'.format(
        jenkins_url, downstream_project, json_request)

    data = json_jenkins.get_json_data(request, append_api=False)
    for build in data['builds']:
        if 'actions' not in build:
            continue
        for action in build['actions']:
            if 'causes' not in action:
                continue
            for cause in action['causes']:
                upstream_project = None
                if isinstance(cause, dict) and 'upstreamProject' in cause:
                    upstream_project = cause['upstreamProject']
                    upstream_build = cause['upstreamBuild']

            if build['building'] and upstream_project == upstream_job:
                #now we need to find the builds of the right upstream_job
                #in data['upstreamProjects']
                builds = {}
                for upstream_project in data['upstreamProjects']:
                    if upstream_project['name'] == upstream_job:
                        builds = upstream_project['builds']

                #find the upstream build in the data
                for upstreamBuild in builds:
                    if upstreamBuild['number'] != upstream_build:
                        continue
                    if _actions_has_all_params(upstreamBuild['actions'],
                                               upstream_params):
                        return True
    return False


def is_job_or_downstream_building(job, job_params):
    """Check if a job (with given parameters) or its downstream job is
    currently building or not.

    Usually this will be called like this:
        is_job_or_downstream_building(
            'my-project-ci',
            {'merge_proposal': 'http://my-merge-proposal/...'})
    It will return True iff:
       - my-project-ci build is running or queued and it has the correct
         merge_proposal param
       - a downstream project (e.g. generic-land) is running (or queued) and
         my-project-ci which triggered the downstream has the right
         merge_proposal param

    This is unfortunately not 100% reliable because is_downstream_job_running
    is not 100% reliable. See the docstring there for more info.

    :param job: name of the jenkins job
    :param job_params: parameters of the jenkins job we want to check
    """

    json_jenkins = get_json_jenkins()
    jenkins_url = get_config_option('jenkins_url')

    # get the jenkins queue. This is to prevent a case when a job is queued
    # while we are preforming the other checks. It might be the other way
    # round (i.e. a job will get out of the queue and finish its execution)
    # and our result won't be correct. Doing it this way is however safer. The
    # worst thing that can happen is that our job will be scheduled with the
    # next poll
    queue = json_jenkins.get_json_data('{}/queue'.format(jenkins_url))

    # first check if a job with a given name and given params isn't already
    # running
    if get_running_builds(job, job_params):
        logger.debug('{} is currently building'.format(job))
        return True

    # next check if such job (including the params) is not queued
    if is_job_queued(job, queue, job_params=job_params):
        logger.debug('{} is currently queued'.format(job))
        return True

    # now do similar checks like the above for the (directly) downstream
    # projects
    downstream_projects = get_downstream_projects(
        json_jenkins, '{}/job/{}'.format(jenkins_url, job))

    for downstream_project in downstream_projects:
        # first check the queue; If the downstream_project is in the queue
        # and was triggered by job then we are done
        if is_job_queued(downstream_project, queue, upstream_job=job,
                         upstream_params=job_params):
            logger.debug('Downstream project "{}" is queued.'.format(
                downstream_project))
            return True

        # build is not queued. Lets check if it is not running
        if is_downstream_job_running(downstream_project, job, job_params):
            logger.debug('Downstream project "{}" is building'.format(
                downstream_project))
            return True

    return False


def is_job_published(job):
    if not job:
        logger.debug('is_job_published: no job specified')
        return False

    jenkins_server = jenkins.Jenkins(get_config_option('jenkins_url'),
                                     get_config_option('jenkins_user'),
                                     get_config_option('jenkins_password'))
    try:
        config = jenkins_server.get_job_config(job)
        dom = parseString(config)
        elements = dom.getElementsByTagName(
            'hudson.plugins.build__publisher.BuildPublisher')
        # if the above element was found then the job is configured to be
        # published. Return true if the length of the array is > 0 and
        # False otherwise
        return elements.length > 0
    except:
        logger.info('Unable to fetch or parse job configuration for: ' +
                    job)
        return False


def hide_jenkins_url(url):
    job = None
    # Match the jenkins JOB_URL and extract the IP/hostname
    # (^http[s]?://.*) and the name of jenkins job.
    # The following format is expected
    # http://10.0.0.1:8080/job/my-jenkins-job/7/
    m = re.match('^(http[s]?://.*)/job/([^/]*)/.*$', url)
    if m:
        jenkins_ip = m.group(1)
        job = m.group(2)
    # if there was a match in the regexp above then both
    # job and jenkins_ip must be set. It is enough to check just for job
    if job and is_job_published(job):
        # if the url ends with "rebuild" its a rebuild URL. In that case
        # we don't want to replace with PUBLIC_JENKINS_URL (which is only a
        # publisher and the jobs are not running there).
        if not url.endswith('rebuild'):
            return url.replace(jenkins_ip,
                               get_config_option('public_jenkins_url'))

    for ip, replacement in get_config_option('urls_to_hide'):
        url = url.replace(ip, replacement)
    return url


def get_multiconfig_builds(jenkins, job_url):
    """ get list of builds for multiconfiguration jobs. The results
        is a list of dictionaries like this:
        {
          'result': 'SUCCESS',
          'output': 'http://jenkins/job/my-job/config_name/43/console'
        }"""

    job = jenkins.get_json_data(job_url + '/')
    data = []
    for run in job['runs']:
        if run['number'] == job['number']:
            result = jenkins.get_json_data(run['url'])['result']
            console_output = normalize_url(run['url'] + 'console')
            data.append({'result': result,
                         'output': console_output})
    return data


def is_matrix_job(jenkins, job_url):
    """ Check if a job on a give job_url is a matrix/multiconfiguration
    """
    #if there is a build number at the end get rid of it
    match = re.match('(^.*/)[0-9]+/?$', job_url)
    if match:
        job_url = match.group(1)
    data = jenkins.get_json_data(job_url)
    return 'activeConfigurations' in data


def get_pipeline_builds(jenkins, job_url):
    """Get a list of stages for pipeline jobs."""
    job = jenkins.get_json_data(job_url + 'wfapi', append_api=False)
    data = []
    for stage in job['stages']:
        data.append({'result': stage['status'],
                     'output': stage['name']})
        if stage['status'] == 'FAILED':
            break

    return data


def is_pipeline_job(jenkins, job_url):
    """ Check if a job is a pipeline job."""
    try:
        data = jenkins.get_json_data(job_url + 'wfapi',
                                     append_api=False)
    except urllib2.HTTPError:
        data = None

    if data:
        return 'stages' in data
    else:
        return False


def parse_jenkins_url(url):
    """ Return a touple of jenkins_url, job_name and build_number given
       a jenkins url """
    build_number_match = re.search('(.*/)(.*)/([0-9]+)/?$', url)
    if build_number_match:
        build_number = int(build_number_match.group(3))
        job_name = build_number_match.group(2)
        jenkins_url = build_number_match.group(1)
        return (jenkins_url, job_name, build_number)
    else:
        return (None, None, None)


def get_executed_builds(jenkins, url):
    """ Find all executed builds no matter if the job is
    multiconfiguration/matrix or if it is triggering downstream jobs

    return value is a list of builds/dictionaries. E.g.
    {'result': 'PASSED',
     'output': 'http://jenkins:8080/job/conf1/console'}
    """

    jenkins_url, job_name, build_number = parse_jenkins_url(url)
    if not jenkins_url:
        return []

    builds = []
    if is_matrix_job(jenkins, jenkins_url + job_name):
        builds = builds + get_multiconfig_builds(jenkins, url)
    elif is_pipeline_job(jenkins, url):
        builds = builds + get_pipeline_builds(jenkins, url)
    else:
        downstream_builds = get_downstream_builds(jenkins,
                                                  jenkins_url,
                                                  job_name,
                                                  build_number)
        for build in downstream_builds:
            if is_matrix_job(jenkins, build['url']):
                builds = builds + get_multiconfig_builds(jenkins,
                                                         build['url'])
            else:
                builds.append({'result': build['result'],
                               'output': normalize_url(
                                   build['url'] + '/' + 'console')})
    return builds


def get_triggered_builds(jenkins, jenkins_url, job_name, build_number):
    """For a given build find all the triggered builds and return them in
       a list where each element is of this form:
       {
       'project':
       'number':
       'url':
       'result':
       }
           """
    return_data = []

    json_request = jenkins_url + job_name + '/' + str(build_number) +\
        '/api/json?depth=2&tree=actions[triggeredBuilds[' +\
        'number,url,result,actions[causes[upstreamBuild,' +\
        'upstreamProject]]]]'
    triggered_builds = jenkins.get_json_data(json_request, append_api=False)

    for action in triggered_builds.get('actions', []):
        for triggered_build in action.get('triggeredBuilds', []):
            causes = {}
            for triggered_action in triggered_build.get('actions', []):
                if 'causes' in triggered_action:
                    causes = triggered_action['causes'][0]
                    break

            # Ensure the build was triggered by the correct upstream project.
            # This avoids including older builds that are erroneously listed
            # by jenkins in some cases (e.g., for rebuilds).
            if (causes.get('upstreamProject', None) != job_name or
                    causes.get('upstreamBuild', None) != build_number):
                continue

            tb_jenkins_url, tb_job_name, tb_build_number = \
                parse_jenkins_url(triggered_build['url'])
            return_data.append({'project': tb_job_name,
                                'number': triggered_build['number'],
                                'url': normalize_url(triggered_build['url']),
                                'result': triggered_build['result']})

    return return_data


def get_downstream_builds(jenkins, jenkins_url, job_name, build_number,
                          depth=0):
    """For a given build find all the downstream builds and return them in
       a list where each element is of this form:
       {
       'project':
       'number':
       'url':
       'result':
       }
           """
    # this is an artifical "if" to stop the recursive calls in case something
    # goes wrong. The conditional should never really evaluate to true (unless
    # you really have job structure 10 levels deep). This mostly serves as a
    # safety net for bugs or wrong assumptions of this code against jenkins
    # behavior
    if depth > 10:
        return []
    job_url = jenkins_url + job_name
    downstream_projects = get_downstream_projects(jenkins, job_url)
    return_data = []

    # Get the triggered builds which in newer versions of the relevant plugin
    # are not included as normal downstream builds
    return_data += \
        get_triggered_builds(jenkins, jenkins_url, job_name, build_number)

    # Add normal downstream builds
    json_request = '/api/json?depth=2&tree=builds[number,url,result,' + \
                   'actions[causes[upstreamBuild,upstreamProject]]]'
    for project in downstream_projects:
        data = jenkins.get_json_data(jenkins_url + project + json_request,
                                     append_api=False)
        for build in data['builds']:
            causes = []
            for action in build['actions']:
                if 'causes' in action:
                    causes = action['causes'][0]
                    break
            upstream_job = None
            upstream_build = None
            if 'upstreamProject' in causes:
                upstream_job = causes['upstreamProject']
            if 'upstreamBuild' in causes:
                upstream_build = causes['upstreamBuild']
            if (upstream_job == job_name and
                    upstream_build == build_number):
                return_data.append({'project': project,
                                   'number': build['number'],
                                   'url': normalize_url(build['url']),
                                   'result': build['result']})

    for job_build in return_data:
        new_data = get_downstream_builds(jenkins, jenkins_url,
                                         job_build['project'],
                                         job_build['number'],
                                         depth=depth + 1)
        return_data = return_data + new_data

    # filter out jobs which are on the blacklist. These are usually something
    # like 'generic-update_mp' but it can be arbitrary job. We are filtering
    # it at the very end as we might want to filter intermediate jobs too
    # (i.e. we might be only interested in results of all the downstream jobs
    # rather than the job itself)
    jobs_blacklisted_from_messages = get_config_option(
        'jobs_blacklisted_from_messages')
    return_data = [p for p in return_data
                   if not p['project'] in jobs_blacklisted_from_messages]

    return _deduplicate_builds(return_data)


def get_downstream_projects(jenkins, job_url):
    data = jenkins.get_json_data(job_url + '/')
    downstream_projects = []

    try:
        for project in data['downstreamProjects']:
            downstream_projects.append(project['name'])
    except KeyError:
        logger.debug('No downstream projects.')

    return downstream_projects


def get_coverity_artifacts(jenkins, url):
    ret = []
    artifacts = jenkins.get_json_data(url)['artifacts']
    for artifact in artifacts:
        path = artifact['relativePath']
        url = url.rstrip('/')
        if path.startswith('results/coverity'):
            link = "{run_url}/artifact/{artifact_path}"
            link = link.format(
                run_url=url,
                artifact_path=path)
            ret.append(normalize_url(link))
    return ret


def get_deb_artifacts(jenkins, url):
    artifacts = jenkins.get_json_data(url)['artifacts']
    for artifact in artifacts:
        path = artifact['relativePath']
        url = url.rstrip('/')
        if path.endswith('.deb'):
            dirname = os.path.dirname(path)
            zipname = os.path.basename(dirname)
            link = "%s/artifact/%s/*zip*/%s.zip" % (url, dirname, zipname)
            return normalize_url(link)
    return None


def get_parameter(jenkins, job_url, name, default=None):
    '''Returns the default value of a given job parameter'''
    data = jenkins.get_json_data(job_url + '/')
    for property_entry in data['property']:
        if 'parameterDefinitions' in property_entry:
            parameter_definitions = property_entry['parameterDefinitions']
            break
    else:
        parameter_definitions = []
    for param in parameter_definitions:
        if param['name'] == name:
            value = param['defaultParameterValue']['value']
            logger.debug("jenkins job parameter [%s]: %s" %
                         (name, value))
            return value
    logger.debug("jenkins job parameter [%s]: not found" % (name))
    if default is None:
        raise KeyError("jenkins job parameter [%s]: not found" % (name))
    logger.debug("Returning default: %s" % (default))
    return default


def get_merge_job(jenkins, job_url):
    '''Retrieves the merge build trigger from a job'''
    merge_jobs = []
    downstream_projects = get_downstream_projects(jenkins, job_url)
    for job_name in downstream_projects:
        m = re.match('(.*-land$)', job_name)
        if m and m not in merge_jobs:
            merge_jobs.append(job_name)
    if len(merge_jobs) == 0:
        raise KeyError("No merge jobs found")
    if len(merge_jobs) > 1:
        raise KeyError("Multiple merge jobs found: %s" %
                       (', '.join(merge_jobs)))
    return merge_jobs[0]


def add_merge_params(jenkins, job_url, params):
    """Populate the jenkins parameters to complete the fasttrack merge"""
    params['target_branch'] = get_parameter(
        jenkins,
        job_url,
        'target_branch')
    params['source_branch'] = params['landing_candidate']
    params['revision'] = params['candidate_revision']
    params['packaging_branch'] = get_parameter(
        jenkins,
        job_url,
        'packaging_branch')
    params['ppa_target'] = get_parameter(jenkins, job_url, 'ppa_target')
    params['distributions'] = get_parameter(jenkins,
                                            job_url,
                                            'distributions')
    params['buildjob_url'] = "Build was skipped, merge only"
    params['use_description_for_commit'] = get_parameter(
        jenkins,
        job_url,
        'use_description_for_commit')
    params['version_string_format'] = get_parameter(
        jenkins,
        job_url,
        'version_string_format')
    params['hooks'] = get_parameter(jenkins, job_url, 'dput_hooks', '')
    params['test_result'] = "PASSED"

def get_jenkins_params(mp):
    params = {
        'landing_candidate': launchpadutils.get_source_repo(mp),
        'merge_proposal': mp.web_link,
        'candidate_revision': launchpadutils.get_latest_revision(mp)}

    if '+git' in mp.web_link:
        params['landing_candidate_branch'] = launchpadutils.get_source_branch(mp)

    return params

def start_jenkins_job(lp_handle, launchpaduser, jenkins_url, jenkins_job, mp,
                      fasttrack_merge=False):
    """Start a given jenkins job on a jenkins as defined in config file

    :param launchpaduser: handle to launchpad user (used to nominate reviewer)
    :param jenkins_url: url of the jenkins instance
    :param jenkins_job: name of the jenkins job to trigger
    :param mp: launchpad merge proposal handle associated with this triggering
    :param fasttrack_merge (optional): boolean deciding whether to use
    fasttrack merge or not if applicable (fasttrack merge only merges the code
    witout actually triggering the testing job; this is applicable only for
    merge proposals modifying the changelog)
    """
    j = jenkins.Jenkins(jenkins_url, get_config_option('jenkins_user'),
                        get_config_option('jenkins_password'))

    jenkins_params = get_jenkins_params(mp)

    job_info = j.get_job_info(jenkins_job)
    if not job_info['buildable']:
        logger.debug('Job is not buildable (probably disabled)')
        return False

    try:
        mp.nominateReviewer(
            reviewer=launchpaduser,
            review_type=get_config_option('launchpad_review_type'))
    except Unauthorized:
        try:
            username = lp_handle.me.name
        except Unauthorized:
            username = '<anonymous>'
        logger.info('You (' + username + ') are not authorized to review '
                    'this merge proposal. CI process will now be stopped.')
        return False

    try:
        is_merge_only = launchpadutils.is_merge_only(mp)
    except LookupError as error:
        logger.error(error.message)
        logger.debug('Skipping this merge proposal')
        return False

    if fasttrack_merge and is_merge_only:
        logger.debug('Skipping the build, merge only')
        json_jenkins = get_json_jenkins()
        job_url = '/'.join([jenkins_url, 'job', jenkins_job]) + '/'
        logger.debug(job_url)
        merge_job = get_merge_job(json_jenkins, job_url)
        add_merge_params(json_jenkins, job_url, jenkins_params)
        logger.debug('Updated params: ' + str(jenkins_params))
        jenkins_job = merge_job
    else:
        logger.debug('Doing a regular build')

    try:
        logger.debug('Starting job (%s): %s' %
                     (jenkins_job, str(jenkins_params)))
        build_token = get_config_option('jenkins_build_token')
        j.build_job(jenkins_job, jenkins_params, build_token)
    except jenkins.JenkinsException:
        logger.debug('Starting a job failed')
        message = "ERROR: failed to start a jenkins job"
        mp.createComment(
            review_type=get_config_option('launchpad_review_type'),
            vote='Disapprove', subject=launchpadutils.get_vote_subject(mp),
            content=message)


def _get_result_line(result, url):
    """Helper method for get_executed_test_runs_message.

    Given a result and url it returns a string.
    In case the result is failed it provides a link to console output:
    FAILED: http://[...]/./distribution=quantal,flavor=amd64/267/console

    In case the result is success or unstable (some tests failed) it provides
    a link to the job:
    SUCCESS: http://[...]/distribution=quantal,flavor=amd64/267

    :param result: one of SUCCESS, UNSTABLE or FAILURE (comes from jenkins)
    :param url: url of the build or configuration in case of matrix job
    """

    url = url.rstrip('/')
    if result == 'SUCCESS' or result == 'UNSTABLE':
        if url.endswith('console'):
            # strip "/console"
            url = url[:-8]
    return "\n    {result}: {output}".format(
        result=result,
        output=url)


def get_executed_test_runs_message(url):
    """Return a string with executed runs for a given jenkins build url.

    :param url: url with jenkins build

    This is an example result of this function:
    ----snip----
Executed test runs:
    FAILURE: http://[...]/./distribution=raring,flavor=amd64/3/console
    FAILURE: http://[...]/./distribution=raring,flavor=i386/3/console
    SUCCESS: http://[...]/job-freestyle/7
        deb: http://[...]/job-freestyle/7/artifact/work/output/*zip*/output.zip
    UNSTABLE: http://jenkins/job/freestyle-downstream/12
Coverity artifacts:
    http://[...]/artifact/results/coverity/CID_10895.html
    http://[...]/results/coverity/CID_10896.html
    http://[...]/results/coverity/CID_10895.html
    http://[...]/results/coverity/CID_10896.html
    ----snip----
    """

    ret = "\nExecuted test runs:"
    jenkins = get_json_jenkins()
    builds = get_executed_builds(jenkins, url)
    coverity_artifacts = []
    for build in builds:
        ret += _get_result_line(
            build['result'],
            hide_jenkins_url(build['output']))
        if build['output'].endswith('console'):
            build_url = build['output'][:-len('console')]
            artifacts = get_coverity_artifacts(jenkins, build_url)
            debs = get_deb_artifacts(jenkins, build_url)
            if debs:
                ret += "\n        deb: {}".format(hide_jenkins_url(debs))
            coverity_artifacts += artifacts
    if coverity_artifacts:
        ret += "\nCoverity artifacts:"
        for artifact in coverity_artifacts:
            ret += "\n    {}".format(hide_jenkins_url(artifact))

    return ret


def format_message_for_mp_update(build_url, message=None,
                                 include_rebuild_link=True):
    """Return a formatted message that is then posted as comment to the merge
    proposal.

    The message consists of:
    * message
    * executed test-runs (and links to artifacts such as deb files)
    * rubuild link

    :params build_url: jenkins url of the build in question
    :params message: custom message (usually PASSED/FAILED)
    :include_rebuild_link: boolean which affects the generation of rebuild link

    Example:
    ----snip----
    PASSED: Continuous integration, rev:114
    http://s-jenkins.ubuntu-ci:8080/job/jenkins-launchpad-plugin-ci/267/
    Executed test runs:
        SUCCESS: http://[...]/distribution=quantal,flavor=amd64/267

    Click here to trigger a rebuild:
    http://s-jenkins.ubuntu-ci:8080/job/jenkins-launchpad-plugin-ci/267/rebuild
    ----snip----

    """
    formatted_message = dedent('''\
            {message}{build_url}{executed_test_runs}
            ''')
    if include_rebuild_link:
            formatted_message = formatted_message + dedent('''\

            Click here to trigger a rebuild:
            {rebuild_url}
            ''')
    if not message:
        message = ''
    if not build_url:
        build_url = ''
        executed_test_runs_message = ''
    else:
        executed_test_runs_message = \
            get_executed_test_runs_message(build_url)

    rebuild_url = '{build_url}/rebuild'.format(
        build_url=build_url.rstrip('/'))
    return formatted_message.format(
        message=message,
        build_url=hide_jenkins_url(build_url),
        rebuild_url=hide_jenkins_url(rebuild_url),
        executed_test_runs=executed_test_runs_message)


def trigger_ci_build(lp_handle, mps, jenkins_job, jenkins_url):
    result = True
    launchpad_user = lp_handle.people(get_config_option('launchpad_login'))
    for mp in mps:

        if not launchpadutils.users_allowed_to_trigger_jobs([mp.registrant]):
            continue

        if launchpadutils.latest_candidate_validated(launchpad_user, mp):
            continue

        if launchpadutils.testing_in_progress(mp, jenkins_job):
            continue

        ret = start_jenkins_job(lp_handle, launchpad_user, jenkins_url,
                                jenkins_job, mp)
        result = ret and result
    return result


def trigger_al_build(lp_handle,
                     mps,
                     jenkins_job,
                     jenkins_url,
                     fasttrack_merge):
    result = True
    for mp in mps:
        #ignore this merge proposal if both registrant and reviewer
        #are not trusted (aka belong to canonical)
        #check if mp.reviewer actually exists just to be sure (it should
        # as we are autolanding and thus the MP needs to be approved)
        if mp.reviewer is None:
            continue
        if not launchpadutils.users_allowed_to_trigger_jobs(
                [mp.registrant, mp.reviewer]):
            logger.debug('Skiping this merge proposal as neither ' +
                         'registrant nor reviewer is allowed to ' +
                         'trigger jenkins jobs')
            continue

        if not launchpadutils.is_commit_message_set(mp,
                                                    jenkins_job,
                                                    jenkins_url):
            logger.debug('No commit message was set. Interrupting ' +
                         'autolanding for "%s".' % (mp.web_link))
            mp_state, message = launchpadutils.build_state({
                'commit_message_empty': True,
                'merge_proposal': mp.web_link})
            launchpadutils.change_mp_status(
                mp, mp_state, message,
                mp.source_branch.revision_count,
                launchpadutils.LaunchpadVote.NEEDS_FIXING)
            continue

        if launchpadutils.unapproved_prerequisite_exists(mp):
            logger.debug('Unapproved prerequisite exists for ' + str(mp))
            continue

        if launchpadutils.testing_in_progress(mp, jenkins_job):
            continue

        logger.debug('Going to trigger merge: ' + str(mp.source_branch))

        launchpad_user = lp_handle.people(
            get_config_option('launchpad_login'))
        ret = start_jenkins_job(lp_handle, launchpad_user, jenkins_url,
                                jenkins_job, mp, fasttrack_merge)
        result = ret and result
    return result
