import jlp.commands.dput as dput
import unittest
from mock import patch, MagicMock


class TestDput(unittest.TestCase):
    def setUp(self):
        self.DputRunner_patch = patch('jlp.commands.dput.DputRunner')
        self.DputRunner = self.DputRunner_patch.start()
        self.minimal_env = {
            'ppa_target': 'ppa',
            'trunk': 'trunk',
            'DEBFULLNAME': 'DEBFULLNAME',
            'DEBEMAIL': 'DEBEMAIL',
            'distributions': 'distributions'
        }

    def tearDown(self):
        self.DputRunner_patch.stop()

    def _get_getenv(self, **kwargs):
        return lambda x: kwargs[x] if x in kwargs else None

    def test_version_string_format_strip(self):
        """If version_string_format is quoted then DputRunner must strip those
        quotes"""
        self.DputRunner.reset_mock()
        version_string = 'version_string'
        my_env = self.minimal_env.copy()
        my_env['version_string_format'] = '"{}"'.format(version_string)
        getenv = self._get_getenv(**my_env)
        with patch('os.getenv', new=getenv):
            ret = dput.dput()
        self.assertEqual(ret, 0)
        self.assertEqual(self.DputRunner.call_args[1]['versionFormat'],
                         version_string)

    def test_multiple_hooks(self):
        """If multiple hooks are specified in environment as csv string then
        DputRunner must be called with a list of hooks"""
        self.DputRunner.reset_mock()
        my_env = self.minimal_env.copy()
        my_env['hooks'] = '1,2,3'
        getenv = self._get_getenv(**my_env)
        with patch('os.getenv', new=getenv):
            ret = dput.dput()
        self.assertEqual(ret, 0)
        self.assertEqual(self.DputRunner.call_args[1]['hooks'],
                         ['1', '2', '3'])

    def test_required_env(self):
        """If any of the parameters defined in self.minimal_env is missing
        then dput command must fail (i.e. return 1) and must not call
        DputRunner"""
        self.DputRunner.reset_mock()
        my_env = self.minimal_env.copy()
        while my_env:
            my_env.popitem()
            getenv = self._get_getenv(**my_env)
            with patch('os.getenv', new=getenv):
                ret = dput.dput()
            self.assertEqual(ret, 1)
            self.assertFalse(self.DputRunner.called)


class TestDputRunnerFailure(unittest.TestCase):
    def test_dputrunner_fails(self):
        """If DputRunner fails then dput must fail too"""
        with patch('os.getenv'), \
                patch('jlp.commands.dput.DputRunner') as DputRunner:
            dput_runner = MagicMock()
            dput_runner.dput = lambda: False
            DputRunner.return_value = dput_runner
            ret = dput.dput()
        self.assertEqual(ret, 1)
        self.assertTrue(DputRunner.called)
