import unittest
from mock import patch, MagicMock
import tests.util as util
from jlp.commands.getMergeProposals import getMergeProposals


class TestGetMergeProposal(unittest.TestCase):

    def setUp(self):
        get_launchpad = lambda launchpadlib_dir: MagicMock()
        self.get_launchpad_patch = patch(
            'jlp.commands.getMergeProposals.get_launchpad',
            new=get_launchpad)
        self.get_launchpad_patch.start()

    def tearDown(self):
        self.get_launchpad_patch.stop()

    def test_gibberish(self):
        """Gibberish on the command line produces usage message."""

        sys_argv = ['--oheck']
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr:
            try:
                getMergeProposals()
            except SystemExit:
                pass
            self.assertTrue("usage" in stderr.getvalue())

    def test_no_branch_specified(self):
        """No branch specified; exit with error message."""

        sys_argv = ['getMergeProposals.py']
        error_message = 'error: too few arguments'
        with patch('sys.argv', sys_argv), util.captured_stderr() as stderr:
            with self.assertRaises(SystemExit) as e:
                getMergeProposals()
            self.assertIn(error_message, stderr.getvalue())
            self.assertEquals(2, e.exception.code)

    def test_no_mp_found_one_branch(self):
        """One branch specified with no MPs found."""

        sys_argv = ['getMergeProposals.py',
                    'lp:~fginther/fake-branch/foo1']
        with patch('sys.argv', sys_argv), \
                patch('jlp.launchpadutils.get_merge_proposals') \
                as get_merge_proposals:
            get_merge_proposals.return_value = []
            count = getMergeProposals()
            self.assertEquals(0, count)

    def test_no_mp_found_multiple_branches(self):
        """Muiltipe branches specified with no MPs found."""

        sys_argv = ['getMergeProposals.py',
                    'lp:~fginther/fake-branch/foo1',
                    'lp:~fginther/fake-branch/foo2',
                    'lp:~fginther/fake-branch/foo3']
        with patch('sys.argv', sys_argv), \
                patch('jlp.launchpadutils.get_merge_proposals') \
                as get_merge_proposals:
            get_merge_proposals.return_value = []
            count = getMergeProposals()
            self.assertEquals(0, count)

    def test_mp_found_one_branch(self):
        """One branch specified with one MP found."""

        sys_argv = ['getMergeProposals.py',
                    'lp:~fginther/fake-branch/foo1']
        with patch('sys.argv', sys_argv), \
                patch('jlp.launchpadutils.get_merge_proposals') \
                as get_merge_proposals:
            get_merge_proposals.return_value = ['Launchpad_mp']
            count = getMergeProposals()
            self.assertEquals(1, count)

    def test_mp_found_multiple_branches(self):
        """One branch specified with multiple MPs found."""

        sys_argv = ['getMergeProposals.py',
                    'lp:~fginther/fake-branch/foo1']
        with patch('sys.argv', sys_argv), \
                patch('jlp.launchpadutils.get_merge_proposals') \
                as get_merge_proposals:
            get_merge_proposals.return_value = ['Launchpad_mp1',
                                                'Launchpad_mp2',
                                                'Launchpad_mp3']
            count = getMergeProposals()
            self.assertEquals(3, count)
