from unittest import TestCase
import jlp
from jlp import (get_json_jenkins, load_config, get_config_option,
                 set_log_level)
from mock import patch, MagicMock
import logging


def custom_open(filename, mode):
    if filename != 'jlp.config':
        raise IOError
    return MagicMock()


class TestGetJsonJenkins(TestCase):
    json_jenkins = object()
    config = object()

    def setUp(self):
        jlp._json_jenkins = None

    @patch('jlp._json_jenkins', new=json_jenkins)
    def test_get_json_jenkins_subsequent_time(self):
        """Existing JSONJenkins class must be returned when calling
        get_json_jenkins 2+ time"""
        ret = get_json_jenkins()
        self.assertEquals(ret, self.json_jenkins)

    @patch('jlp.load_config')
    def test_get_json_replace_no_url(self, load_config):
        """Ensure the url is unchanged when no jenkins_url is defined."""
        load_config.return_value = {
            'jenkins_url': '',
            'jenkins_proxy_url': 'https://proxy/',
            'jenkins_user': 'user',
            'jenkins_password': 'password'
        }
        json_jenkins = get_json_jenkins()
        url = json_jenkins.patch_url('https://proxy/job/my-job')
        self.assertEquals(url, 'https://proxy/job/my-job')

    @patch('jlp.load_config')
    def test_get_json_replace_no_proxy_url(self, load_config):
        """Ensure the url is unchanged when no proxy URL is defined."""
        load_config.return_value = {
            'jenkins_url': 'https://backend:8080/',
            'jenkins_proxy_url': '',
            'jenkins_user': 'user',
            'jenkins_password': 'password'
        }
        json_jenkins = get_json_jenkins()
        url = json_jenkins.patch_url('https://proxy/subdir/job/my-job')
        self.assertEquals(url, 'https://proxy/subdir/job/my-job')

    @patch('jlp.load_config')
    def test_get_json_replace_proxy_url(self, load_config):
        """Ensure that the proxy url gets replaced if both the jenkins_url
        and jenkins_proxy_url are defined."""
        load_config.return_value = {
            'jenkins_url': 'http://backend:8080/',
            'jenkins_proxy_url': 'https://proxy/',
            'jenkins_user': 'user',
            'jenkins_password': 'password'
        }
        json_jenkins = get_json_jenkins()
        url = json_jenkins.patch_url('https://proxy/subdir/job/my-job')
        self.assertEquals(url, 'http://backend:8080/subdir/job/my-job')

    @patch('jlp._json_jenkins', new=None)
    @patch('jlp.JSONJenkins')
    def test_get_json_jenkins_first_time(self, json_jenkins):
        """Test that json_jenkins is returned when get_json_jenkins is called
        for the first time"""
        ret = get_json_jenkins()
        self.assertEquals(json_jenkins.call_count, 1)
        self.assertIsNotNone(ret)

    @patch('jlp._config', new=config)
    @patch('jlp.yaml.safe_load')
    def test_load_config_subsequent_time(self, safe_load):
        ret = load_config()
        self.assertEquals(safe_load.call_count, 0)
        self.assertEquals(ret, self.config)

    @patch('jlp._config', new=None)
    @patch('__builtin__.open')
    @patch('jlp.yaml.safe_load')
    def test_load_config_from_home_first_time(self, safe_load, open_mock):
        ret = load_config()
        self.assertEquals(safe_load.call_count, 1)
        self.assertIsNotNone(ret)

    @patch('jlp._config', new=None)
    @patch('__builtin__.open', new=custom_open)
    @patch('jlp.yaml.safe_load')
    def test_load_config_from_cwd_first_time(self, safe_load):
        ret = load_config()
        self.assertIsNotNone(ret)
        self.assertEquals(safe_load.call_count, 1)

    @patch('jlp._config', new=None)
    @patch('__builtin__.open')
    def test_no_config_found(self, open_mock):
        """If no config is found jlp should sys.exit with 1"""
        open_mock.side_effect = IOError()
        systemExit = False
        try:
            load_config()
        except SystemExit as e:
            self.assertEquals(e.code, 1)
            systemExit = True
        self.assertTrue(systemExit)

    @patch('jlp.load_config', new=lambda: {'my_option': 'my_value'})
    def test_get_config_option(self):
        """get_config_option must return the right value for a defined
        option"""
        value = get_config_option('my_option')
        self.assertEquals(value, 'my_value')

    @patch('jlp.load_config', new=lambda: {'my_option': 'my_value'})
    def test_get_invalid_config_option(self):
        """ get_config_option must raise KeyError when option does not
        exist"""
        keyError = False
        try:
            get_config_option('invalid_option')
        except KeyError:
            keyError = True
        self.assertTrue(keyError)

    @patch('jlp.stdout_handler')
    @patch('jlp.logger')
    @patch('jlp.get_config_option')
    def test_no_log_level(self, get_config_option, logger, stdout_handler):
        """If no log_level is specified then it must default to
        logging.DEBUG"""
        get_config_option.side_effect = KeyError()
        set_log_level(logger, stdout_handler)
        logger.setLevel.assert_called_once_with(logging.DEBUG)
        stdout_handler.setLevel.assert_called_once_with(logging.DEBUG)

    @patch('jlp.stdout_handler')
    @patch('jlp.logger')
    @patch('jlp.get_config_option')
    def test_set_log_level(self, get_config_option, logger, stdout_handler):
        """If no log_level is set to 'info' it must be set to logging.INFO"""
        get_config_option.return_value = 'info'
        set_log_level(logger, stdout_handler)
        logger.setLevel.assert_called_once_with(logging.INFO)
        stdout_handler.setLevel.assert_called_once_with(logging.INFO)
