import unittest
from mock import patch, MagicMock
import tests.util as util
from jlp.commands.launchpadTrigger import launchpadTrigger, trigger_jenkins
from jlp import get_config_option
import os


class PatchedLaunchpadTriggerTest(unittest.TestCase):

    def setUp(self):
        self.launchpad = MagicMock()
        get_launchpad = lambda x: self.launchpad
        self.get_launchpad_patch = patch(
            'jlp.commands.launchpadTrigger.get_launchpad', new=get_launchpad)
        self.get_launchpad_patch.start()

    def tearDown(self):
        self.get_launchpad_patch.stop()


class TestLaunchpadTriggerParams(PatchedLaunchpadTriggerTest):

    merge_proposal = 'https://code.launchpad.net/~mrazik/' + \
                     'faux-dbus-test-runner/yar8/+merge/106113'

    def test_gibberish(self):
        """Gibberish on the command line produces usage message."""

        sys_argv = ['--oheck']
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr:
            try:
                launchpadTrigger()
            except SystemExit:
                pass
            self.assertTrue("usage" in stderr.getvalue())

    def test_no_trigger_ci_nor_autoland_specified(self):
        """Generate the 'specify either -t or -l' error."""

        sys_argv = ['launchpadTrigger.py', '-b', 'lp:faux-dbus-test-runner',
                    '-j', 'test-job', '-e', 'jenkins-url']
        error_message = 'You need to specify one of --trigger-ci (-t)' + \
                        ' or --autoland (-l)'
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr, \
                patch('sys.exit') as sys_exit:
            launchpadTrigger()
            sys_exit.assert_called_with(2)
            self.assertTrue(error_message in stderr.getvalue())

    def test_no_branch_specified(self):
        """Failed to specify branch produces error."""

        sys_argv = ["launchpadTrigger.py",
                    "-a",
                    "--job=faux-dbus-test-runner",
                    "--autoland"]
        error_message = "error: argument -b/--branch is required"
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr:
            try:
                launchpadTrigger()
            except SystemExit:
                pass
            self.assertTrue(error_message in stderr.getvalue())

    def test_trigger_autolanding_no_jenkins_job(self):
        """Attempt to trigger fails with no Jenkins job."""

        sys_argv = ["launchpadTrigger.py",
                    "-a",
                    "-n",
                    "--branch=lp:faux-dbus-test-runner",
                    "--autoland"]
        error_message = 'argument -j/--job is required'
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr:
            try:
                launchpadTrigger()
            except SystemExit:
                pass
            self.assertTrue(error_message in stderr.getvalue())

    def test_trigger_autolanding(self):
        """Trigger autolanding with a set of command-line params."""

        sys_argv = ["launchpadTrigger.py",
                    "--job=faux-dbus-test-runner",
                    "--branch=lp:faux-dbus-test-runner",
                    "--autoland"]
        with patch('sys.argv', sys_argv), \
                patch('jlp.commands.launchpadTrigger.get_merge_proposals') as \
                get_merge_proposals, \
                patch('jlp.commands.launchpadTrigger.jenkinsutils.' +
                      'trigger_al_build') as trigger_al_build:
            get_merge_proposals.return_value = [self.merge_proposal]
            launchpadTrigger()
            trigger_al_build.assert_called_with(self.launchpad,
                                                [self.merge_proposal],
                                                'faux-dbus-test-runner',
                                                'http://localhost:8080/',
                                                False)

    def test_trigger_ci(self):
        """Trigger CI with a set of command-line params."""

        sys_argv = ["launchpadTrigger.py",
                    "--branch=lp:faux-dbus-test-runner",
                    "--trigger-ci",
                    "--job=faux-dbus-test-runner"]
        with patch('sys.argv', sys_argv), \
                patch('jlp.commands.launchpadTrigger.get_merge_proposals') as \
                get_merge_proposals, \
                patch('jlp.commands.launchpadTrigger.jenkinsutils.' +
                      'trigger_ci_build') as trigger_ci_build:
            get_merge_proposals.return_value = [self.merge_proposal]
            launchpadTrigger()
            trigger_ci_build.assert_called_with(self.launchpad,
                                                [self.merge_proposal],
                                                'faux-dbus-test-runner',
                                                'http://localhost:8080/')

    def test_trigger_ci_needs_review(self):
        """Trigger CI with a set of command-line params."""

        sys_argv = ["launchpadTrigger.py",
                    "--branch=lp:faux-dbus-test-runner",
                    "--trigger-ci",
                    "--job=faux-dbus-test-runner"]
        with patch('sys.argv', sys_argv), \
                patch('jlp.commands.launchpadTrigger.get_merge_proposals') as \
                get_merge_proposals, \
                patch('jlp.commands.launchpadTrigger.jenkinsutils.' +
                      'trigger_ci_build') as trigger_ci_build:
            get_merge_proposals.return_value = [self.merge_proposal]
            launchpadTrigger()
            trigger_ci_build.assert_called_with(self.launchpad,
                                                [self.merge_proposal],
                                                'faux-dbus-test-runner',
                                                'http://localhost:8080/')


    def test_trigger_git_ci(self):
        """Trigger CI got a git repo."""

        sys_argv = ["launchpadTrigger.py",
                    "--branch=lp:faux-dbus-test-runner",
                    "--trigger-ci",
                    "--job=faux-dbus-test-runner",
                    "--repo_type=git"]
        with patch('sys.argv', sys_argv), \
             patch('jlp.commands.launchpadTrigger.get_merge_proposals') as \
                get_merge_proposals, \
                patch('jlp.commands.launchpadTrigger.jenkinsutils.' +
                              'trigger_ci_build') as trigger_ci_build:
            get_merge_proposals.return_value = [self.merge_proposal]
            launchpadTrigger()
            get_merge_proposals.assert_called_with(self.launchpad,
                                                   'lp:faux-dbus-test-runner',
                                                   ['Needs review'],
                                                   'git')


class TestLockNames(PatchedLaunchpadTriggerTest):

    sys_argv = ["launchpadTrigger.py",
                "--job=faux-dbus-test-runner",
                "--branch=lp:faux-dbus-test-runner",
                "--autoland"]

    def setUp(self):
        super(TestLockNames, self).setUp()
        # Ensure that the fake branch names don't also map to git repos
        self.launchpad.git_repositories = MagicMock()
        self.launchpad.git_repositories.getByPath = MagicMock()
        self.launchpad.git_repositories.getByPath.return_value = None

    def tearDown(self):
        super(TestLockNames, self).tearDown()

    def test_trigger_without_lock_name(self):
        """Trigger autolanding with a default launchpadlib lockname."""

        with patch('sys.argv', self.sys_argv), \
                patch('jlp.commands.launchpadTrigger.SocketLock') \
                as socketLock:
            launchpadTrigger()
            socketLock.assert_called_with(
                get_config_option('lock_name') + '-launchpadTrigger')

    def test_trigger_with_lock_name(self):
        """Trigger autolanding with a custom launchpadlib lockname."""
        lock_name = 'qa.cfg'

        with patch('sys.argv', self.sys_argv + ['-L', lock_name]), \
                patch('jlp.commands.launchpadTrigger.SocketLock') \
                as socketLock:
            launchpadTrigger()
            socketLock.assert_called_with(
                get_config_option('lock_name') + '-' + lock_name)


class TestTriggerJenkins(unittest.TestCase):

    lock_name = 'qa.cfg'
    expected_cache_dir = os.path.expanduser('~/.launchpadlib-qa.cfg')

    def test_lock_name_ci(self):
        """test trigger_jenkins method for correct use of lock_name param"""

        with patch('jlp.commands.launchpadTrigger.trigger_ci_build') \
                as trigger_ci_build:
            trigger_jenkins(branch=None,
                            trigger_ci=True,
                            lock_name=self.lock_name)
            trigger_ci_build.assert_called_with(
                None, None, get_config_option('jenkins_url'),
                self.expected_cache_dir, 'auto')

    def test_lock_name_autolanding(self):
        """test trigger_jenkins method for correct use of lock_name param"""

        with patch('jlp.commands.launchpadTrigger.trigger_al_build') \
                as trigger_al_build:
            trigger_jenkins(branch=None,
                            autoland=True,
                            lock_name=self.lock_name)
            trigger_al_build.assert_called_with(
                None, None, get_config_option('jenkins_url'),
                False, self.expected_cache_dir, 'auto')

    def test_repo_type_is_propagated(self):
        """test trigger_jenkins method for correct use of lock_name param"""

        with patch('jlp.commands.launchpadTrigger.trigger_al_build') \
                as trigger_al_build:
            trigger_jenkins(branch=None,
                            autoland=True,
                            lock_name=self.lock_name,
                            repo_type='git')
            trigger_al_build.assert_called_with(
                None, None, get_config_option('jenkins_url'),
                False, self.expected_cache_dir, 'git')
