import unittest
from jlp import launchpadutils, get_config_option
from mock import MagicMock, patch
from testscenarios import TestWithScenarios
from textwrap import dedent


class LaunchpadUtilsTests(unittest.TestCase):
    launchpadAgent = None

    def test_get_branch_from_mp(self):
        branch = launchpadutils.get_branch_from_mp(
            'https://copde.launchpad.net/~user/project/name/+merge/12345')
        expected = 'https://copde.launchpad.net/~user/project/name/'
        self.assertEqual(branch, expected)

    def test_get_branch_from_mp_invalid_url(self):
        branch = launchpadutils.get_branch_from_mp(
            'https://copde.launchpad.net/~user/project/name/merge/12345')
        self.assertEqual(branch, None)

    @patch('jlp.launchpadutils.get_vote_subject',
           new=lambda mp: 'Subject')
    def test_change_mp_status_with_no_vote(self):
        mp = MagicMock()
        launchpadutils.change_mp_status(mp, 'Status', 'Message', revision=1)
        mp.createComment.assert_called_once_with(
            review_type=get_config_option('launchpad_review_type'),
            subject='Subject',
            content='Message')
        mp.setStatus.assert_called_once_with(status='Status', revid=1)

    @patch('jlp.launchpadutils.get_vote_subject',
           new=lambda mp: 'Subject')
    def test_change_mp_status_with_approve(self):
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        launchpadutils.change_mp_status(mp, 'Status', 'Message',
                                        revision=1, vote='Approve')
        mp.createComment.assert_called_once_with(
            review_type=get_config_option('launchpad_review_type'),
            subject='Subject',
            content='Message', vote='Approve')
        mp.setStatus.assert_called_once_with(status='Status', revid=1)

    @patch('jlp.launchpadutils.get_vote_subject',
           new=lambda mp: 'Subject')
    def test_change_mp_status_with_disapprove(self):
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        launchpadutils.change_mp_status(mp, 'Status', 'Message',
                                        revision=1, vote='Disapprove')
        mp.createComment.assert_called_once_with(
            review_type=get_config_option('launchpad_review_type'),
            subject='Subject',
            content='Message', vote='Disapprove')
        mp.setStatus.assert_called_once_with(status='Status', revid=1)

    def test_close_bugs(self):
        mp = MagicMock()
        mp.target_branch.project.web_link = 'http://launchpad/jenkins'
        mp.target_branch.project.name = 'project'
        mp.source_branch.linked_bugs_collection = [MagicMock(), MagicMock()]
        mp.target_branch.display_name = 'lp:project'
        mp.merged_revno = '12'
        mp.target_branch.self_link = 'http://project/project'
        milestone = MagicMock()
        milestone.series_target.branch_link = mp.target_branch.self_link
        milestone.name = '5.0'
        mp.target_branch.project.active_milestones_collection = [milestone]
        for b in mp.source_branch.linked_bugs_collection:
            b.bug_tasks_collection = [MagicMock(), MagicMock(), MagicMock()]
            bug_task = b.bug_tasks_collection[1]
            bug_task.target.web_link = mp.target_branch.project.web_link

        launchpadutils.close_bugs(mp)
        for b in mp.source_branch.linked_bugs_collection:
            b.bug_tasks_collection[1].lp_save.assert_called_once_with()
            b.newMessage.assert_called_once_with(
                content='Fix committed into lp:project at revision ' +
                '12, scheduled for release in project, milestone 5.0')

    def test_get_reviews_no_reviews(self):
        mp = MagicMock()
        mp.votes = []
        self.assertTrue(launchpadutils.get_reviews(mp) is None)

    def test_get_reviews(self):
        mp = MagicMock()
        mp.votes = [MagicMock(), MagicMock(), MagicMock()]
        for vote in mp.votes:
            vote.reviewer.display_name = 'jenkins'
            vote.comment.vote = 'Approve'
        reviews = launchpadutils.get_reviews(mp)
        for review in reviews:
            self.assertEquals(review, 'jenkins;Approve')

    def test_unapproved_prerequisite_no_prerequisite(self):
        mp = MagicMock()
        mp.prerequisite_branch = None
        self.assertFalse(launchpadutils.unapproved_prerequisite_exists(mp))

    def test_unapproved_prerequisite_too_many_proposals(self):
        mp = MagicMock()
        mp.target_branch = MagicMock()
        mp.target_branch.web_link = 'url'
        prereq = MagicMock()
        prereq.landing_targets = [MagicMock(), MagicMock(), MagicMock()]
        for t in prereq.landing_targets:
                t.target_branch.web_link = 'url'
                t.queue_status = 'Approved'
        mp.prerequisite_branch = prereq
        self.assertTrue(launchpadutils.unapproved_prerequisite_exists(mp))

    def test_unapproved_prerequisite_no_proposal_for_prerequisite(self):
        mp = MagicMock()
        prereq = MagicMock()
        prereq.landing_targets = []
        mp.prerequisite_branch = prereq
        self.assertTrue(launchpadutils.unapproved_prerequisite_exists(mp))

    def test_unapproved_prerequisite_not_merged_yet(self):
        mp = MagicMock()
        mp.target_branch = MagicMock()
        mp.target_branch.web_link = 'url'
        prereq = MagicMock()
        prereq.landing_targets = [MagicMock()]
        for t in prereq.landing_targets:
                t.target_branch.web_link = 'url'
                t.queue_status = 'Approved'
        mp.prerequisite_branch = prereq
        self.assertTrue(launchpadutils.unapproved_prerequisite_exists(mp))

    def test_unapproved_prerequisite_merged(self):
        mp = MagicMock()
        mp.target_branch = MagicMock()
        mp.target_branch.web_link = 'url'
        prereq = MagicMock()
        prereq.landing_targets = [MagicMock()]
        for t in prereq.landing_targets:
                t.target_branch.web_link = 'url'
                t.queue_status = 'Merged'
        mp.prerequisite_branch = prereq
        self.assertFalse(launchpadutils.unapproved_prerequisite_exists(mp))


class TestUserAllowedToTriggerJobs(unittest.TestCase):

    def test_user_not_allowed_to_trigger_job(self):
        registrant = MagicMock()
        registrant.name = 'random-user'
        membership = MagicMock()
        membership.team.name = 'random-team'
        registrant.memberships_details = [membership]
        result = launchpadutils.users_allowed_to_trigger_jobs([registrant])
        self.assertFalse(result)

    def test_user_member_of_allowed_team(self):
        registrant = MagicMock()
        registrant.name = 'random-user'
        team = MagicMock()
        team.participants = [registrant]
        registrant._root.people = {'ps-jenkins': team}
        result = launchpadutils.users_allowed_to_trigger_jobs([registrant])
        self.assertTrue(result)

    def test_users_allowed_to_trigger(self):
        registrant = MagicMock()
        registrant.name = 'ps-jenkins'
        membership = MagicMock()
        membership.team.name = ''
        registrant.memberships_details = [membership]
        result = launchpadutils.users_allowed_to_trigger_jobs([registrant])
        self.assertTrue(result)

    def test_users_allowed_to_trigger_no_users(self):
        self.assertFalse(launchpadutils.users_allowed_to_trigger_jobs([]))

    def test_allow_indirect_membership(self):
        registrant = MagicMock()
        registrant.name = 'random-user'
        team = MagicMock()
        team.is_team = True
        registrant._root.people = {'ps-jenkins': team}
        with patch('jlp.launchpadutils.users_in_team') as mock:
            launchpadutils.users_allowed_to_trigger_jobs([registrant])
        mock.assert_called_with([registrant], team)

    def test_skip_non_team(self):
        registrant = MagicMock()
        registrant.name = 'random-user'
        person = MagicMock()
        person.is_team = False
        registrant._root.people = {'ps-jenkins': person}
        with patch('jlp.launchpadutils.users_in_team') as mock:
            launchpadutils.users_allowed_to_trigger_jobs([registrant])
        self.assertEqual(0, mock.call_count)

    def test_users_in_team(self):

        class Team(object):

            def __init__(self, participants):
                self._participants = participants

            @property
            def participants(self):
                return iter(self._participants)

        present = object()
        absent = object()
        team = Team([present])
        # We want the actual boolean values, not values that evaluate to True
        # or False.
        self.assertIs(True, launchpadutils.users_in_team([present], team))
        self.assertIs(False, launchpadutils.users_in_team([absent], team))
        self.assertTrue(launchpadutils.users_in_team([absent, present], team))

    def test_allow_not_an_lp_user(self):
        """Verify that a missing lp user exception is handled."""
        registrant = MagicMock()
        registrant.name = 'random-user'
        # Simulate a missing lp user by
        # returning a dict that has no
        # users
        registrant._root.people = {}
        result = launchpadutils.users_allowed_to_trigger_jobs([registrant])
        self.assertFalse(result)


class TestIsMergeOnly(unittest.TestCase):
    def test_is_merge_only(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = MagicMock()
        diff_stat = {u'debian/changelog': 'diff text'}
        mp.preview_diff.diffstat = diff_stat
        self.assertTrue(launchpadutils.is_merge_only(mp))

    def test_is_merge_only_not_changelog(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = MagicMock()
        diff_stat = {u'debian/rules': 'diff text'}
        mp.preview_diff.diffstat = diff_stat
        self.assertFalse(launchpadutils.is_merge_only(mp))

    def test_is_merge_only_multiple_files(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = MagicMock()
        diff_stat = {u'debian/changelog': 'diff text',
                     u'debian/rules': 'diff text'}
        mp.preview_diff.diffstat = diff_stat
        self.assertFalse(launchpadutils.is_merge_only(mp))

    def test_is_merge_only_no_preview_diff(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = None
        with self.assertRaises(LookupError):
            launchpadutils.is_merge_only(mp)

    def test_is_merge_only_no_diffstat(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = MagicMock()
        mp.preview_diff.diffstat = None
        with self.assertRaises(LookupError):
            launchpadutils.is_merge_only(mp)


class TestGetCommitMessage(TestWithScenarios):
    scenarios = [
        ('no_commit_message',
         {'commit_message': None,
          'description': 'description',
          'use_description': False,
          'expected': None}),
        ('no_commit_message_fallback_to_description',
         {'commit_message': None,
          'description': 'description',
          'use_description': True,
          'expected': 'description'}),
        ('commit_message_and_description',
         {'commit_message': 'commit_message',
          'description': 'description',
          'use_description': True,
          'expected': 'commit_message'}),
        ('no_commit_message_and_no_description',
         {'commit_message': '',
          'description': '',
          'use_description': True,
          'expected': None})]

    def test_get_commit_message(self):
        mp = MagicMock()
        mp.commit_message = self.commit_message
        mp.description = self.description
        use_description = self.use_description

        message = launchpadutils.get_commit_message(mp, use_description)
        self.assertEqual(message, self.expected)


class LaunchpadTriggerIsCommitMessageSet(unittest.TestCase):
    def _setup_json_data(self, json_data):
        jenkins = MagicMock()
        jenkins.get_json_data = MagicMock(return_value=json_data)
        self.get_json_jenkins_patch = patch(
            'jlp.launchpadutils.get_json_jenkins',
            new=MagicMock(return_value=jenkins))
        self.get_json_jenkins_mock = self.get_json_jenkins_patch.start()

    def _setup_get_commit_message(self, return_value):
        get_commit_message = MagicMock(return_value=return_value)
        self.get_commit_message_patch = patch(
            'jlp.launchpadutils.get_commit_message',
            new=get_commit_message)
        self.get_commit_message_mock = self.get_commit_message_patch.start()

    def _cleanup(self):
        self.get_commit_message_patch.stop()
        self.get_json_jenkins_patch.stop()

    def test_use_description_but_still_no_commit_message(self):
        json_data = {'property': ['dummy', {'parameterDefinitions': [{
            'name': 'use_description_for_commit',
            'defaultParameterValue': {'value': True}}]}]}
        self._setup_json_data(json_data)
        self._setup_get_commit_message(False)
        ret = launchpadutils.is_commit_message_set(None, 'job', 'url')
        self.get_commit_message_mock.assert_called_once_with(None, True)
        self.assertFalse(ret)
        self._cleanup()

    def test_no_use_description_but_commit_message_set(self):
        json_data = {'property': ['dummy', {'parameterDefinitions': [{
            'name': 'use_description_for_commit',
            'defaultParameterValue': {'value': False}}]}]}
        self._setup_json_data(json_data)
        self._setup_get_commit_message(True)
        ret = launchpadutils.is_commit_message_set(None, 'job', 'url')
        self.get_commit_message_mock.assert_called_once_with(None, False)
        self.assertTrue(ret)
        self._cleanup()

    def test_use_description_flag_not_available(self):
        json_data = {}
        self._setup_json_data(json_data)
        self._setup_get_commit_message(True)
        ret = launchpadutils.is_commit_message_set(None, 'job', 'url')
        self.get_commit_message_mock.assert_called_once_with(None, False)
        self.assertTrue(ret)
        self._cleanup()

    @patch('jlp.launchpadutils.jenkinsutils.is_job_or_downstream_building',
           new=lambda jenkins_job, job_params: False)
    def test_testing_not_in_progress_approved(self):
        """ If there is no job running then testing_in_progress
        must return False.
        """
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        self.assertFalse(launchpadutils.testing_in_progress(mp, 'job'))

    @patch('jlp.launchpadutils.jenkinsutils.is_job_or_downstream_building',
           new=lambda jenkins_job, job_params: True)
    def test_testing_not_in_progress_abstain(self):
        """ If there is job running
        then testing_in_progress must return True and it must report the
        correct jenkins job which holds the lock.
        """
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        self.assertTrue(launchpadutils.testing_in_progress(mp, 'job'))


class TestLatestReview(unittest.TestCase):
    def setUp(self):
        self.launchpad_user = MagicMock()
        self.jenkinsUserName = 'name'
        self.launchpad_user.name = self.jenkinsUserName

    def test_get_latest_review_no_comments(self):
        mp = MagicMock()
        mp.all_comments = []
        self.assertEqual(
            launchpadutils.get_latest_review(self.launchpad_user, mp), 0)

    def test_get_latest_review_unrelated_comment(self):
        mp = MagicMock()
        comment = MagicMock()
        comment.vote_tag = get_config_option('launchpad_review_type')
        comment.message_body = 'unrelated comment'
        mp.all_comments = [comment]
        self.assertEqual(
            launchpadutils.get_latest_review(self.launchpad_user, mp), 0)

    def test_get_latest_review_which_passed(self):
        mp = MagicMock()
        comment = MagicMock()
        comment.author.name = self.jenkinsUserName
        comment.vote_tag = get_config_option('launchpad_review_type')
        comment.message_body = "PASSED: Continuous integration, rev:47\n" + \
                               "URL:..."
        mp.all_comments = [comment]
        self.assertEqual(
            launchpadutils.get_latest_review(self.launchpad_user, mp), '47')

    def test_get_latest_review_which_failed(self):
        mp = MagicMock()
        comment = MagicMock()
        comment.author.name = self.jenkinsUserName
        comment.vote_tag = get_config_option('launchpad_review_type')
        comment.message_body = "FAILED: Continuous integration, rev:47\n" + \
                               "URL:..."
        mp.all_comments = [comment]
        self.assertEqual(
            launchpadutils.get_latest_review(self.launchpad_user, mp), '47')

    @patch('jlp.launchpadutils.get_latest_review', new=lambda x, y: 60)
    def test_new_mp_candidate_revision_in_queue(self):
            #60 is the revision which was validated as last
            mp = MagicMock()
            #this is the current branch revision
            mp.source_branch.revision_count = 61
            #latest_candidate_validated should return False as there is a new
            #revision in the queue (61)
            ret = launchpadutils.latest_candidate_validated(
                self.launchpad_user, mp)
            self.assertFalse(ret)

    @patch('jlp.launchpadutils.get_latest_review', new=lambda x, y: 60)
    def test_mp_already_validated(self):
            #60 is the revision which was validated as last
            mp = MagicMock()
            #this is the current branch revision
            mp.source_branch.revision_count = 60
            #latest_candidate_validated should return True. This rev was
            #already validated
            ret = launchpadutils.latest_candidate_validated(
                self.launchpad_user, mp)
            self.assertTrue(ret)

class TestLatestReviewForGitMPs(unittest.TestCase):
    def setUp(self):
        self.launchpad_user = MagicMock()
        self.jenkinsUserName = 'name'
        self.launchpad_user.name = self.jenkinsUserName

    def create_mp(self):
        mp = MagicMock()
        mp.web_link = 'https://code.launchpad.net/~user/project/+git/project/+merge/777'
        return mp

    def create_mp_with_revision(self, revision):
        mp = self.create_mp()
        path = "refs/heads/nice"
        ref = MagicMock()
        ref.path = path
        ref.commit_sha1 = revision
        mp.source_git_path = path
        mp.source_git_repository.refs_collection = [ref]
        return mp

    def test_get_latest_review_no_comments(self):
        mp = self.create_mp()
        mp.all_comments = []
        self.assertEqual(
            launchpadutils.get_latest_review(self.launchpad_user, mp), 0)

    def test_get_latest_review_unrelated_comment(self):
        mp = self.create_mp()
        comment = MagicMock()
        comment.vote_tag = get_config_option('launchpad_review_type')
        comment.message_body = 'unrelated comment'
        mp.all_comments = [comment]
        self.assertEqual(
            launchpadutils.get_latest_review(self.launchpad_user, mp), 0)

    def test_get_latest_review_which_passed(self):
        mp = self.create_mp()
        comment = MagicMock()
        comment.author.name = self.jenkinsUserName
        comment.vote_tag = get_config_option('launchpad_review_type')
        comment.message_body = "PASSED: Continuous integration, rev:2fd4e1c67a2d28fced849ee1bb76e7391b93eb12\n" + \
                               "URL:..."
        mp.all_comments = [comment]
        self.assertEqual(
            launchpadutils.get_latest_review(self.launchpad_user, mp), '2fd4e1c67a2d28fced849ee1bb76e7391b93eb12')

    def test_get_latest_review_which_failed(self):
        mp = self.create_mp()
        comment = MagicMock()
        comment.author.name = self.jenkinsUserName
        comment.vote_tag = get_config_option('launchpad_review_type')
        comment.message_body = "FAILED: Continuous integration, rev:2fd4e1c67a2d28fced849ee1bb76e7391b93eb12\n" + \
                               "URL:..."
        mp.all_comments = [comment]
        self.assertEqual(
            launchpadutils.get_latest_review(self.launchpad_user, mp), '2fd4e1c67a2d28fced849ee1bb76e7391b93eb12')

    @patch('jlp.launchpadutils.get_latest_review', new=lambda x, y: '2fd4e1c67a2d28fced849ee1bb76e7391b93eb12')
    def test_new_mp_candidate_revision_in_queue(self):
        mp = self.create_mp_with_revision('de9f2c7fd25e1b3afad3e85a0bd17d9b100db4b3')
        ret = launchpadutils.latest_candidate_validated(
            self.launchpad_user, mp)
        self.assertFalse(ret)

    @patch('jlp.launchpadutils.get_latest_review', new=lambda x, y: '2fd4e1c67a2d28fced849ee1bb76e7391b93eb12')
    def test_mp_already_validated(self):
        mp = self.create_mp_with_revision('2fd4e1c67a2d28fced849ee1bb76e7391b93eb12')
        ret = launchpadutils.latest_candidate_validated(
            self.launchpad_user, mp)
        self.assertTrue(ret)

class TestApproveAndDisapprove(unittest.TestCase):
    jenkinsUserName = 'jenkins-user'

    def _get_mp(self):
        mp = MagicMock()
        mp.web_link = 'http://mp'
        mp.target_branch.display_name = 'lp:unity'
        mp.source_branch.display_name = 'lp:~ps-jenkins/unity/fix'
        return mp

    def _get_git_mp(self):
        mp = MagicMock()
        mp.web_link = 'https://code.launchpad.net/~dummy/agitproject/+git/gitproject/+merge/778899'
        mp.target_git_repository.display_name = 'lp:agitproject'
        mp.source_git_repository.display_name = 'lp:~auser/gitrepo'
        mp.source_git_path = 'refs/heads/abranch'
        mp.target_git_path = 'refs/heads/master'
        return mp

    def _get_subject(self, mp):
        return 'Re: [Merge] {} into {}'.format(
            mp.source_branch.display_name, mp.target_branch.display_name)

    def _expected_git_mp_subject(self, mp):
        source = mp.source_git_repository.display_name.replace('lp:', '') + \
                 ':' + \
                 mp.source_git_path.replace('refs/heads/', '')
        target = mp.target_git_repository.display_name.replace('lp:', '') + \
                 ':' + \
                 mp.target_git_path.replace('refs/heads/', '')

        return 'Re: [Merge] {} into {}'.format(source, target)

    def setUp(self):
        user = MagicMock()
        user.name = self.jenkinsUserName
        self.get_executed_test_runs_message_patch = patch(
            'jlp.jenkinsutils.get_executed_test_runs_message',
            new=lambda x: '')
        self.get_executed_test_runs_message_patch.start()

    def tearDown(self):
        self.get_executed_test_runs_message_patch.stop()

    def test_approve_mp(self):
        mp = self._get_mp()
        subject = self._get_subject(mp)
        launchpadutils.approve_mp(mp, '1', 'http://url')
        content = dedent('''\
                            PASSED: Continuous integration, rev:1
                            http://url

                            Click here to trigger a rebuild:
                            http://url/rebuild
                            ''')
        mp.createComment.assert_called_once_with(
            review_type=get_config_option('launchpad_review_type'),
            vote=launchpadutils.LaunchpadVote.APPROVE,
            subject=subject,
            content=content)

    def test_disapprove_mp(self):
        mp = self._get_mp()
        subject = self._get_subject(mp)
        launchpadutils.disapprove_mp(mp, '1', 'http://url')
        content = dedent('''\
                            FAILED: Continuous integration, rev:1
                            http://url

                            Click here to trigger a rebuild:
                            http://url/rebuild
                            ''')
        mp.createComment.assert_called_once_with(
            review_type=get_config_option('launchpad_review_type'),
            vote=launchpadutils.LaunchpadVote.NEEDS_FIXING,
            subject=subject,
            content=content)

    def test_disapprove_mp_with_reason(self):
        mp = self._get_mp()
        subject = self._get_subject(mp)
        reason = "No commit message set\nDo something\n"
        content = dedent('''\
                            FAILED: Continuous integration, rev:1
                            {reason}
                            http://url

                            Click here to trigger a rebuild:
                            http://url/rebuild
                            ''').format(reason=reason)

        launchpadutils.disapprove_mp(mp,
                                     '1',
                                     'http://url',
                                     reason=reason)
        mp.createComment.assert_called_once_with(
            review_type=get_config_option('launchpad_review_type'),
            vote=launchpadutils.LaunchpadVote.NEEDS_FIXING,
            subject=subject,
            content=content)

    def test_approve_git_mp(self):
        mp = self._get_git_mp()
        subject = self._expected_git_mp_subject(mp)
        launchpadutils.approve_mp(mp, '1', 'http://url')
        content = dedent('''\
                            PASSED: Continuous integration, rev:1
                            http://url

                            Click here to trigger a rebuild:
                            http://url/rebuild
                            ''')
        mp.createComment.assert_called_once_with(
            review_type=get_config_option('launchpad_review_type'),
            vote=launchpadutils.LaunchpadVote.APPROVE,
            subject=subject,
            content=content)

    def test_dissaprove_git_mp(self):
        mp = self._get_git_mp()
        subject = self._expected_git_mp_subject(mp)
        launchpadutils.disapprove_mp(mp, '1', 'http://url')
        content = dedent('''\
                            FAILED: Continuous integration, rev:1
                            http://url

                            Click here to trigger a rebuild:
                            http://url/rebuild
                            ''')
        mp.createComment.assert_called_once_with(
            review_type=get_config_option('launchpad_review_type'),
            vote=launchpadutils.LaunchpadVote.NEEDS_FIXING,
            subject=subject,
            content=content)


class TestGetMergeProposalsWithInvalidBranch(unittest.TestCase):
    branch_with_mps = None

    def setUp(self):
        branch = MagicMock()
        branch.landing_candidates = []
        for i in range(1, 10):
            mp = MagicMock()
            mp.queue_status = i
            branch.landing_candidates.append(mp)
        self.branch_with_mps = branch

    def _get_launchpad(self, getByUrl_result=None):
        launchpad = MagicMock()
        launchpad.branches.getByUrl = lambda url: getByUrl_result
        return launchpad

    def test_get_merge_proposals_with_invalid_branch(self):
        launchpad = self._get_launchpad(None)
        ret = launchpadutils.get_merge_proposals(launchpad, '', '', 'bazaar')
        self.assertEqual(ret, [])

    def test_get_merge_proposals_odd(self):
        launchpad = self._get_launchpad(self.branch_with_mps)
        mps = launchpadutils.get_merge_proposals(launchpad, '', [1, 3, 5, 7],
                                                 'bazaar')
        self.assertEqual(len(mps), 4)

    def test_get_merge_proposal_two(self):
        launchpad = self._get_launchpad(self.branch_with_mps)
        mps = launchpadutils.get_merge_proposals(launchpad, '', [2], 'bazaar')
        self.assertEqual(len(mps), 1)
        self.assertEqual(mps[0].queue_status, 2)

class TestGetGitMergeProposals(unittest.TestCase):
    git_repo_with_mps = None
    mp_status = range(1,10)

    def setUp(self):
        repo = MagicMock()
        repo.landing_candidates = []
        for i in self.mp_status:
            mp = MagicMock()
            mp.queue_status = i
            repo.landing_candidates.append(mp)
        self.git_repo_with_mps = repo

    def _get_launchpad(self, getByPath_result=None):
        launchpad = MagicMock()
        launchpad.git_repositories.getByPath = lambda path: getByPath_result
        return launchpad

    def raise_attribute_error(self):
        raise AttributeError

    def _get_launchpad_without_git(self):
        launchpad = MagicMock()
        launchpad.git_repositories = lambda path: self.raise_attribute_error
        return launchpad

    def test_get_git_merge_proposals(self):
        launchpad = self._get_launchpad(self.git_repo_with_mps)
        mps = launchpadutils.get_merge_proposals(launchpad, '', self.mp_status, repo_type='git')
        self.assertEqual(len(mps), len(self.mp_status))
        for mp, expected_status in zip(mps, self.mp_status):
            self.assertEqual(mp.queue_status, expected_status)

    def test_get_git_merge_proposals_with_invalid_branch(self):
        launchpad = self._get_launchpad(None)
        ret = launchpadutils.get_merge_proposals(launchpad, '', '', repo_type='git')
        self.assertEqual(ret, [])

    def test_get_git_merge_proposals_with_wrong_launchpad_version(self):
        launchpad = self._get_launchpad_without_git()
        with self.assertRaises(RuntimeError):
            ret = launchpadutils.get_merge_proposals(launchpad, '', '',
                                                     repo_type='git')

    def test_invalid_repo_type(self):
        launchpad = self._get_launchpad(None)
        with self.assertRaises(RuntimeError):
            ret = launchpadutils.get_merge_proposals(launchpad, '', '',
                                                     repo_type='test')

class TestGetMpHandleFromURL(unittest.TestCase):
    @patch("jlp.launchpadutils.get_branch_from_mp", new=lambda x: None)
    def test_unable_to_get_branch_link(self):
        """If there is no way how to get branch link from the merge proposal
        url then get_mp_handle_from_url must return None"""
        ret = launchpadutils.get_mp_handle_from_url(MagicMock(), 'http://')
        self.assertEquals(ret, None)

    @patch("jlp.launchpadutils.get_branch_from_mp", new=lambda x: 'http://')
    def test_unable_to_get_branch_handle(self):
        """getByURL launchpad call returns None then get_mp_handle_from_url
        must return None too"""
        launchpad = MagicMock()
        launchpad.branches.getByUrl = lambda url: None
        ret = launchpadutils.get_mp_handle_from_url(launchpad, 'http://')
        self.assertEquals(ret, None)

    @patch("jlp.launchpadutils.get_branch_from_mp", new=lambda x: 'http://')
    def test_get_mp_handle_from_url(self):
        """If there is a merge proposal for the given merge proposal URL then
        get_mp_handle_from_url must return the respective merge proposal
        handle"""

        mp_url = 'http://my-url'
        launchpad = MagicMock()
        branch = MagicMock
        mp = MagicMock()
        mp.web_link = mp_url
        branch.landing_targets = [mp]
        launchpad.branches.getByUrl = lambda url: branch
        ret = launchpadutils.get_mp_handle_from_url(launchpad, mp_url)
        self.assertEquals(ret, mp)

    def test_call_git_api_for_git_url(self):
        """If there is a merge proposal for the given merge proposal URL then
        get_mp_handle_from_url must return the respective merge proposal
        handle"""

        mp_url = 'https://code.launchpad.net/~dummy/agitproject/+git/gitproject/+merge/778899'
        launchpad = MagicMock()
        git_repo = MagicMock
        mp = MagicMock()
        mp.web_link = mp_url
        git_repo.landing_targets = [mp]
        launchpad.git_repositories.getByPath = lambda path: git_repo
        ret = launchpadutils.get_mp_handle_from_url(launchpad, mp_url)
        self.assertEquals(ret, mp)

class TestGetBranchHandle(unittest.TestCase):
    def setUp(self):
        '''Setup an empty launchpad.'''
        self.launchpad = MagicMock()

        # Setup for no bazaar branches matched
        self.getByUrlMock = MagicMock()
        self.getByUrlMock.return_value = None
        self.launchpad.branches.getByUrl = self.getByUrlMock

        # Setup for no git repos matched
        self.getByPathMock = MagicMock()
        self.getByPathMock.return_value = None
        self.launchpad.git_repositories.getByPath = self.getByPathMock

    def create_fake_bazaar_branch(self):
        '''Create a matching bazaar branch.'''
        dummy_handle = MagicMock()
        self.getByUrlMock.return_value = dummy_handle
        return dummy_handle

    def create_fake_git_repo(self):
        '''Create a matching git repo.'''
        dummy_handle = MagicMock()
        self.getByPathMock.return_value = dummy_handle
        return dummy_handle

    def test_uses_bzr_api_for_bzr_type(self):
        '''
        Ensure that a bazaar type and branch are returned for bazaar type.
        '''
        name = 'lp:dummy'
        branch = self.create_fake_bazaar_branch()
        result = launchpadutils.get_branch_handle_for(self.launchpad, name,
                                                      'bazaar');

        self.launchpad.branches.getByUrl.assert_called_once_with(url=name)
        self.assertEqual(result, ('bazaar', branch))

    def test_uses_git_api_for_git_type(self):
        '''
        Ensure that a git type and repo are returned for a git type.
        '''
        name = 'lp:dummy'
        repo = self.create_fake_git_repo()
        result = launchpadutils.get_branch_handle_for(self.launchpad, name,
                                                      'git');

        self.launchpad.git_repositories.getByPath.assert_called_once_with(
            path='dummy')
        self.assertEqual(result, ('git', repo))

    def test_detects_bzr_for_bzr_branch(self):
        '''
        Ensure that bazaar is returned for a single match.
        '''
        name = 'lp:dummy'
        branch = self.create_fake_bazaar_branch()
        result = launchpadutils.get_branch_handle_for(self.launchpad, name,
                                                      'auto');

        self.launchpad.branches.getByUrl.assert_called_once_with(url=name)
        self.assertEqual(result, ('bazaar', branch))

    def test_detects_git_for_git_repo(self):
        '''
        Ensure that git is returned for a single match.
        '''
        name = 'lp:dummy'
        repo = self.create_fake_git_repo()
        result = launchpadutils.get_branch_handle_for(self.launchpad, name,
                                                      'auto');

        self.launchpad.git_repositories.getByPath.assert_called_once_with(
            path='dummy')
        self.assertEqual(result, ('git', repo))

    def test_raises_for_ambiguous_branch(self):
        '''
        Ensure that an exception is raised for a url with multiple matches.
        '''
        name = 'lp:dummy'
        self.create_fake_bazaar_branch()
        self.create_fake_git_repo()
        with self.assertRaises(RuntimeError):
            launchpadutils.get_branch_handle_for(self.launchpad, name,
                                                 'auto');

        self.launchpad.branches.getByUrl.assert_called_once_with(url=name)
        self.launchpad.git_repositories.getByPath.assert_called_once_with(
            path='dummy')

    def test_returns_branch_for_branch_and_repo(self):
        '''
        Ensure requested bazaar type for a url with multiple matches.
        '''
        name = 'lp:dummy'
        # Create multiple matches to be returned for the url
        branch = self.create_fake_bazaar_branch()
        repo = self.create_fake_git_repo()
        result = launchpadutils.get_branch_handle_for(self.launchpad, name,
                                                      'bazaar');
        self.assertEqual(result, ('bazaar', branch))

    def test_returns_repo_for_branch_and_repo(self):
        '''
        Ensure requested git type for a url with multiple matches.
        '''
        name = 'lp:dummy'
        # Create multiple matches to be returned for the url
        branch = self.create_fake_bazaar_branch()
        repo = self.create_fake_git_repo()
        result = launchpadutils.get_branch_handle_for(self.launchpad, name,
                                                      'git');

        self.assertEqual(result, ('git', repo))

    def test_raises_for_no_mathc(self):
        '''
        Ensure that an exception is raised when there is no match.
        '''
        name = 'lp:dummy'
        # Use the empty setup state which returns no bazaar or git matches
        with self.assertRaises(RuntimeError):
            launchpadutils.get_branch_handle_for(self.launchpad, name,
                                                 'auto');
