import unittest
from mock import patch, MagicMock
import tests.util as util
from jlp.commands.voteOnMergeProposal import voteOnMergeProposal


class TestVoteOnMergeProposal(unittest.TestCase):

    def setUp(self):
        get_launchpad = lambda launchpadlib_dir: MagicMock()
        self.get_launchpad_patch = patch(
            'jlp.commands.voteOnMergeProposal.get_launchpad',
            new=get_launchpad)
        self.get_launchpad_patch.start()
        self.get_executed_test_runs_message_patch = patch(
            'jlp.jenkinsutils.get_executed_test_runs_message')
        self.get_executed_test_runs_message_patch.start()
        self.is_commit_message_set_patch = patch(
            'jlp.launchpadutils.is_commit_message_set')
        self.is_commit_message_set_patch.start()

    def tearDown(self):
        self.get_launchpad_patch.stop()
        self.is_commit_message_set_patch.stop()
        self.get_executed_test_runs_message_patch.stop()

    def test_gibberish(self):
        """Gibberish on the command line produces usage message."""

        sys_argv = ['--oheck']
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr:
            try:
                voteOnMergeProposal()
            except SystemExit:
                pass
            self.assertTrue("usage" in stderr.getvalue())

    def test_no_build_url_specified(self):
        """No build_url specified; exit with message."""

        sys_argv = ['voteOnMergeProposal.py',
                    '--status=PASSED',
                    '--revision=66',
                    '--branch="lp:~mrazik/faux-dbus-test-runner/yar8"',
                    '--merge-proposal="https://code.launchpad.net/~mrazik/' +
                    'faux-dbus-test-runner/yar8/+merge/106113"']
        error_message = 'error: argument -u/--build-url is required'
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr:
            with self.assertRaises(SystemExit) as e:
                voteOnMergeProposal()
            self.assertTrue(error_message in stderr.getvalue())
            self.assertEquals(2, e.exception.code)

    def test_no_revision_specified(self):
        """No revision specified; exit with message."""

        sys_argv = ['voteOnMergeProposal.py',
                    '--status=PASSED',
                    '--build-url="http://s-jenkins.ubuntu-ci/job/' +
                    'faux-dbus-test-runner-ci/configure"',
                    '--branch="lp:~mrazik/faux-dbus-test-runner/yar8"',
                    '--merge-proposal="https://code.launchpad.net/~mrazik/' +
                    'faux-dbus-test-runner/yar8/+merge/106113"']
        error_message = 'error: argument -r/--revision is required'
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr:
            with self.assertRaises(SystemExit) as e:
                voteOnMergeProposal()
            self.assertTrue(error_message in stderr.getvalue())
            self.assertEquals(2, e.exception.code)

    def test_no_merge_proposal_specified(self):
        """No merge_proposal specified; exit with message."""

        sys_argv = ['voteOnMergeProposal.py',
                    '--status=PASSED',
                    '--build-url="http://s-jenkins.ubuntu-ci/job/' +
                    'faux-dbus-test-runner-ci/configure"',
                    '--revision=66',
                    '--branch="lp:~mrazik/faux-dbus-test-runner/yar8"']
        error_message = 'error: argument -p/--merge-proposal is required'
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr:
            with self.assertRaises(SystemExit) as e:
                voteOnMergeProposal()
            self.assertTrue(error_message in stderr.getvalue())
            self.assertEquals(2, e.exception.code)

    def test_unknown_merge_proposal(self):
        """The given merge proposal wasn't found; error out with message."""

        sys_argv = ['voteOnMergeProposal.py',
                    '--status=PASSED',
                    '--build-url="http://s-jenkins.ubuntu-ci/job/' +
                    'faux-dbus-test-runner-ci/configure"',
                    '--revision=66',
                    '--branch="lp:~mrazik/faux-dbus-test-runner/yar8"',
                    '--merge-proposal=https://code.launchpad.net/~mrazik/' +
                    'faux-dbus-test-runner/yar8/+merge/106113']
        error_message = 'error: merge proposal related to this branch was ' + \
                        'not found'
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr, \
                patch('jlp.commands.voteOnMergeProposal.launchpadutils.' +
                      'get_mp_handle_from_url') as get_mp_handle_from_url:
            get_mp_handle_from_url.return_value = None
            try:
                voteOnMergeProposal()
            except SystemExit:
                pass
            self.assertTrue(error_message in stderr.getvalue())

    def test_passed_vote(self):
        """Vote on a proposal which passed."""

        sys_argv = ['voteOnMergeProposal.py',
                    '--status=PASSED',
                    '--build-url=http://s-jenkins.ubuntu-ci/job/' +
                    'faux-dbus-test-runner-ci/configure',
                    '--revision=66',
                    '--branch=lp:~mrazik/faux-dbus-test-runner/yar8',
                    '--merge-proposal=https://code.launchpad.net/~mrazik/' +
                    'faux-dbus-test-runner/yar8/+merge/106113']
        with patch('sys.argv', sys_argv), \
                patch('jlp.commands.voteOnMergeProposal.launchpadutils.' +
                      'get_mp_handle_from_url') as get_mp_handle_from_url, \
                patch('jlp.launchpadutils.approve_mp') as approve_mp:
            get_mp_handle_from_url.return_value = 'Launchpad_mp'
            ret = voteOnMergeProposal()
            self.assertEqual(ret, 0)
            approve_mp.assert_called_with(
                'Launchpad_mp',
                '66',
                'http://s-jenkins.ubuntu-ci/job/faux-dbus-test-runner-ci/' +
                'configure')

    def test_no_commit_message_set(self):
        """Vote on a proposal where tests passed but no commit message
           was set."""

        merge_proposal = 'https://code.launchpad.net/~mrazik/' + \
                         'faux-dbus-test-runner/yar8/+merge/106113'
        sys_argv = ['voteOnMergeProposal.py',
                    '--status=PASSED',
                    '--build-url=http://s-jenkins.ubuntu-ci/job/' +
                    'faux-dbus-test-runner-ci/configure',
                    '--revision=66',
                    '--branch=lp:~mrazik/faux-dbus-test-runner/yar8',
                    '--merge-proposal=' + merge_proposal]
        with patch('sys.argv', sys_argv), \
                patch('jlp.commands.voteOnMergeProposal.launchpadutils.' +
                      'get_mp_handle_from_url') as get_mp_handle_from_url, \
                patch('jlp.launchpadutils.is_commit_message_set') \
                as is_commit_message_set, \
                patch('jlp.launchpadutils.disapprove_mp') as disapprove_mp:
            get_mp_handle_from_url.return_value = 'Launchpad_mp'
            is_commit_message_set.return_value = False
            reason = "No commit message was specified in the merge " + \
                     "proposal. Click on the following link and set the " + \
                     "commit message (if you want a jenkins rebuild you " +\
                     "need to trigger it yourself):\n" + \
                     merge_proposal + "/+edit-commit-message\n"
            ret = voteOnMergeProposal()
            self.assertEqual(ret, 0)
            disapprove_mp.assert_called_with(
                'Launchpad_mp',
                '66',
                'http://s-jenkins.ubuntu-ci/job/faux-dbus-test-runner-ci/' +
                'configure',
                reason)

    def test_invalid_build_url(self):
        """Vote on a proposal where tests passed but no commit message
           was set."""

        sys_argv = ['voteOnMergeProposal.py',
                    '--status=PASSED',
                    '--build-url=http://s-jenkins.ubuntu-ci/asdf/' +
                    'faux-dbus-test-runner-ci/configure',
                    '--revision=66',
                    '--branch=lp:~mrazik/faux-dbus-test-runner/yar8',
                    '--merge-proposal=https://code.launchpad.net/~mrazik/' +
                    'faux-dbus-test-runner/yar8/+merge/106113']
        with patch('sys.argv', sys_argv), \
                patch('jlp.commands.voteOnMergeProposal.launchpadutils.' +
                      'get_mp_handle_from_url') as get_mp_handle_from_url, \
                patch('jlp.commands.voteOnMergeProposal.launchpadutils.' +
                      'is_commit_message_set') as is_commit_message_set, \
                patch('jlp.launchpadutils.approve_mp') as approve_mp:
            get_mp_handle_from_url.return_value = 'Launchpad_mp'
            is_commit_message_set.return_value = False
            ret = voteOnMergeProposal()
            self.assertEqual(ret, 0)
            approve_mp.assert_called_with(
                'Launchpad_mp',
                '66',
                'http://s-jenkins.ubuntu-ci/asdf/faux-dbus-test-runner-ci/' +
                'configure')

    def test_not_passed_vote(self):
        """Vote on a proposal which did not pass."""

        sys_argv = ['voteOnMergeProposal.py',
                    '--status=OHECK',
                    '--build-url=http://s-jenkins.ubuntu-ci/job/' +
                    'faux-dbus-test-runner-ci/configure',
                    '--revision=66',
                    '--branch="lp:~mrazik/faux-dbus-test-runner/yar8"',
                    '--merge-proposal=https://code.launchpad.net/~mrazik/' +
                    'faux-dbus-test-runner/yar8/+merge/106113']
        with patch('sys.argv', sys_argv), \
                patch('jlp.commands.voteOnMergeProposal.launchpadutils.' +
                      'get_mp_handle_from_url') as get_mp_handle_from_url, \
                patch('jlp.launchpadutils.disapprove_mp') as disapprove_mp:
            get_mp_handle_from_url.return_value = 'Launchpad_mp'
            ret = voteOnMergeProposal()
            self.assertEqual(ret, 0)
            disapprove_mp.assert_called_with(
                'Launchpad_mp',
                '66',
                'http://s-jenkins.ubuntu-ci/job/faux-dbus-test-runner-ci/' +
                'configure',
                '')
